# -*- coding: utf-8 -*-
"""
Test the Angle reparameterisation
"""
import numpy as np
import pytest
from unittest.mock import MagicMock, create_autospec

from nessai.reparameterisations import Angle
from nessai.livepoint import (
    empty_structured_array,
    numpy_array_to_live_points,
    parameters_to_live_point,
)
from nessai.utils.testing import assert_structured_arrays_equal

scales = [1.0, 2.0]


@pytest.fixture(params=scales, scope="function")
def scale(request):
    return request.param


@pytest.fixture
def reparam():
    return create_autospec(Angle)


@pytest.fixture(scope="function")
def assert_invertibility(model, n=100):
    def test_invertibility(reparam, angle, radial=None):

        x = empty_structured_array(n, names=reparam.parameters)
        x_prime = empty_structured_array(n, names=reparam.prime_parameters)
        log_j = 0

        x[reparam.angle] = angle
        if radial is not None:
            x[reparam.radial] = radial

        assert x.size == x_prime.size

        x_re, x_prime_re, log_j_re = reparam.reparameterise(x, x_prime, log_j)

        np.testing.assert_array_equal(x[reparam.angle], x_re[reparam.angle])
        if radial is not None:
            np.testing.assert_array_equal(
                x[reparam.radial], x_re[reparam.radial]
            )

        x_in = empty_structured_array(n, names=reparam.parameters)

        x_inv, x_prime_inv, log_j_inv = reparam.inverse_reparameterise(
            x_in, x_prime_re, log_j
        )

        np.testing.assert_array_almost_equal(
            x[reparam.angle], x_inv[reparam.angle]
        )
        if radial is not None:
            np.testing.assert_array_almost_equal(
                x[reparam.radial], x_inv[reparam.radial]
            )

        assert_structured_arrays_equal(x_prime_re, x_prime_inv)
        np.testing.assert_array_almost_equal(log_j_re, -log_j_inv)

        return True

    return test_invertibility


@pytest.mark.parametrize(
    "bounds, scale, expected_scale",
    [
        ([0, 2 * np.pi], None, 1.0),
        ([-1, 1], None, np.pi),
        ([0, np.pi], 1.0, 1.0),
    ],
)
def test_angle_parameter(bounds, scale, expected_scale):
    """Test init with just an angle parameter"""
    parameter = "theta"
    prior_bounds = {parameter: bounds}
    reparam = Angle(
        parameters=parameter, prior_bounds=prior_bounds, scale=scale
    )

    assert reparam.chi is not False
    assert hasattr(reparam.chi, "rvs")
    if bounds[0] == 0.0:
        assert reparam._zero_bound is True
    else:
        assert reparam._zero_bound is False
    assert reparam.has_prime_prior is False

    assert reparam.angle == parameter
    assert reparam.radial == (parameter + "_radial")
    assert reparam.radius == (parameter + "_radial")
    assert reparam.x == (parameter + "_x")
    assert reparam.y == (parameter + "_y")
    assert reparam.scale == expected_scale


def test_angle_too_many_parameters(reparam):
    """Assert an error is raised if too many parameters are given."""
    parameters = ["x", "y", "z"]
    prior_bounds = {p: [-1, 1] for p in parameters}
    with pytest.raises(RuntimeError) as excinfo:
        Angle.__init__(
            reparam, parameters=parameters, prior_bounds=prior_bounds
        )
    assert reparam.parameters == parameters
    assert "Too many parameters for Angle" in str(excinfo.value)


def test_angle_prior_uniform():
    """Assert the prior is correctly assigned for uniform prior"""
    from nessai.priors import log_2d_cartesian_prior

    reparam = Angle(
        parameters="theta",
        prior="uniform",
        prior_bounds={"theta": [0, np.pi]},
        scale=2.0,
    )
    assert reparam.prior == "uniform"
    assert reparam.has_prime_prior is True
    assert reparam._k == (2 * np.pi)
    assert reparam._prime_prior is log_2d_cartesian_prior


def test_angle_prior_sine():
    """Assert the prior is correctly assigned for sine prior"""
    from nessai.priors import log_2d_cartesian_prior_sine

    reparam = Angle(
        parameters="theta",
        prior="sine",
        prior_bounds={"theta": [0, np.pi]},
    )
    assert reparam.prior == "sine"
    assert reparam.has_prime_prior is True
    assert reparam._k == np.pi
    assert reparam._prime_prior is log_2d_cartesian_prior_sine


def test_log_prior(reparam):
    """Assert the log-prior calls the correct function"""
    reparam.chi = MagicMock()
    reparam.chi.logpdf = MagicMock()
    reparam.parameters = ["theta", "theta_radial"]
    x = {"theta": [1.0], "theta_radial": [0.5]}
    Angle.log_prior(reparam, x)
    reparam.chi.logpdf.assert_called_once_with([0.5])


def test_x_prime_log_prior(reparam):
    """ "Assert the underlying functions is called correctly"""
    x_prime = numpy_array_to_live_points(np.array([[1.0, -1.0]]), ["x", "y"])

    reparam._k = 0.5
    reparam._prime_prior = MagicMock(return_value=0.5)
    reparam.has_prior_prior = True
    reparam.prime_parameters = ["x", "y"]

    out = Angle.x_prime_log_prior(reparam, x_prime)

    reparam._prime_prior.assert_called_once_with(
        x_prime["x"],
        x_prime["y"],
        k=0.5,
    )
    assert out == 0.5


def test_x_prime_log_prior_error(reparam):
    """
    Assert an error is raised when called the prime prior if is not enabled.
    """
    reparam.has_prime_prior = False
    x = {"theta": [1.0], "theta_radial": [0.5]}
    with pytest.raises(RuntimeError) as excinfo:
        Angle.x_prime_log_prior(reparam, x)
    assert "Prime prior" in str(excinfo.value)


def test_both_parameters():
    """Test init with just an angle and radial parameter"""
    parameters = ["theta", "r"]
    prior_bounds = {
        parameters[0]: np.array([0, 2 * np.pi]),
        parameters[1]: np.array([0, 5]),
    }

    reparam = Angle(parameters=parameters, prior_bounds=prior_bounds)

    assert reparam.chi is False
    assert reparam._zero_bound is True
    assert reparam.has_prime_prior is False

    assert reparam.angle == parameters[0]
    assert reparam.radial == parameters[1]


@pytest.mark.parametrize(
    "angle_prior", [np.array([0, 2 * np.pi]), np.array([-np.pi, np.pi])]
)
def test_invertiblity_single_parameter(
    angle_prior, scale, assert_invertibility
):
    """Test the inverbility when using just an angle"""
    n = 100
    parameter = "theta"
    prior_bounds = {parameter: angle_prior / scale}
    reparam = Angle(
        parameters=parameter, prior_bounds=prior_bounds, scale=scale
    )
    angle = np.random.uniform(*prior_bounds[parameter], n)
    assert assert_invertibility(reparam, angle, radial=None)


@pytest.mark.parametrize(
    "angle_prior", [np.array([0, 2 * np.pi]), np.array([-np.pi, np.pi])]
)
def test_invertiblity_both_parameters(
    angle_prior, scale, assert_invertibility
):
    """Test the inverbility when using just an angle"""
    n = 100
    parameters = ["theta", "r"]
    prior_bounds = {
        parameters[0]: angle_prior / scale,
        parameters[1]: np.array([0, 5]),
    }
    reparam = Angle(
        parameters=parameters, prior_bounds=prior_bounds, scale=scale
    )
    angle = np.random.uniform(*prior_bounds[parameters[0]], n)
    radial = np.random.uniform(*prior_bounds[parameters[1]], n)
    assert assert_invertibility(reparam, angle, radial=radial)


@pytest.mark.integration_test
@pytest.mark.parametrize(
    "value, output_x, output_y",
    [
        (-1.0, None, 0.0),
        (1.0, None, 0.0),
        (-0.5, 0.0, None),
        (0.5, 0.0, None),
    ],
)
def test_periodic_parameter(value, output_x, output_y):
    """Test a generic periodic parameter"""
    parameters = ["a"]
    prior_bounds = {"a": [-1.0, 1.0]}
    reparam = Angle(
        parameters=parameters, prior_bounds=prior_bounds, scale=None
    )
    x = parameters_to_live_point((value,), parameters)
    x_prime = parameters_to_live_point(
        (np.nan, np.nan), reparam.prime_parameters
    )
    log_j = np.zeros(x.size)
    x_out, x_prime_out, log_j_out = reparam.reparameterise(x, x_prime, log_j)

    assert_structured_arrays_equal(x_out, x)

    if output_x:
        assert x_prime_out[reparam.x] == output_x
    if output_y:
        assert x_prime_out[reparam.y] == output_y


def test_reparameterise_negative_radius(reparam):
    """Assert an error is radius if the radius is negative."""
    x = numpy_array_to_live_points(
        np.array([[1.0, -1.0]]), ["theta", "radius"]
    )
    x_prime = x.copy()
    log_j = 0.0
    reparam.radial = "radius"
    reparam.chi = None

    reparam._rescale_angle = lambda *args: (x["theta"], *args)
    reparam._rescale_radial = lambda *args: (x["radius"], *args)

    with pytest.raises(RuntimeError) as excinfo:
        Angle.reparameterise(reparam, x, x_prime, log_j)
    assert "Radius cannot be negative" in str(excinfo.value)



# -*- coding: utf-8 -*-
"""
Deprecated submodule that will be removed in a future release.
Use `nessai.samplers.nestedsampler` instead.
"""
from warnings import warn

from .samplers.nestedsampler import NestedSampler  # noqa

msg = (
    "`nessai.nestedsampler` is deprecated and will be removed in a future "
    "release. Use `nessai.samplers.nestedsampler` instead."
)
warn(msg, FutureWarning)



# -*- coding: utf-8 -*-
"""Specific tests for different included flows."""
import numpy as np
import pytest
import torch

from nessai.flows import (
    MaskedAutoregressiveFlow,
    NeuralSplineFlow,
    RealNVP,
)


@pytest.mark.parametrize(
    "kwargs",
    [
        dict(
            net="mlp", batch_norm_within_layers=True, dropout_probability=0.5
        ),
        dict(use_volume_preserving=True),
        dict(actnorm=True),
        dict(linear_transform="permutation"),
        dict(linear_transform="svd"),
        dict(linear_transform="lu"),
        dict(linear_transform=None),
        dict(mask=np.array([[1, -1], [-1, 1]])),
        dict(mask=[1, -1]),
        dict(pre_transform="batch_norm"),
        dict(pre_transform="batch_norm", pre_transform_kwargs=dict(eps=1e-8)),
        dict(scale_activation=lambda x: torch.sigmoid(x + 2) + 1e-3),
    ],
)
def test_with_realnvp_kwargs(kwargs):
    """Test RealNVP with specific kwargs"""
    flow = RealNVP(2, 2, 2, 2, **kwargs)
    x = torch.randn(10, 2)
    z, _ = flow.forward(x)
    assert z.shape == (10, 2)


@pytest.mark.parametrize(
    "kwargs, string",
    [
        (dict(net="res"), "Unknown nn type: res"),
        (dict(linear_transform="test"), "Unknown linear transform: test"),
        (dict(mask=[1, 1, -1]), "Mask does not match number of features"),
        (
            dict(mask=[[-1, 1], [1, -1], [1, -1]]),
            "Mask does not match number of layers",
        ),
    ],
)
def test_realnvp_value_errors(kwargs, string):
    """Assert incorrect values for some inputs raise an error"""
    with pytest.raises(ValueError) as excinfo:
        RealNVP(2, 2, 2, 2, **kwargs)
    assert string in str(excinfo.value)


def test_realnvp_actnorm_batchnorm():
    """
    Assert an error is raised if actnorm and batchnorm are enabled at once.
    """
    with pytest.raises(
        RuntimeError, match=r"Cannot enable actnorm and batchnorm .*"
    ):
        RealNVP(2, 2, 2, 2, actnorm=True, batch_norm_between_layers=True)


@pytest.mark.parametrize(
    "kwargs",
    [
        dict(batch_norm_between_layers=True),
        dict(linear_transform="permutation"),
        dict(linear_transform="svd"),
        dict(linear_transform="lu"),
        dict(linear_transform=None),
        dict(num_bins=10),
    ],
)
def test_with_nsf_kwargs(kwargs):
    """Test NSF with specific kwargs"""
    flow = NeuralSplineFlow(2, 2, 2, 2, **kwargs)
    x = torch.randn(10, 2)
    z, _ = flow.forward(x)
    assert z.shape == (10, 2)


@pytest.mark.parametrize(
    "kwargs",
    [
        dict(batch_norm_between_layers=True),
        dict(batch_norm_within_layers=True),
        dict(use_random_permutations=True),
        dict(use_residual_blocks=True),
        dict(use_random_masks=False),
    ],
)
def test_with_maf_kwargs(kwargs, caplog):
    """Test MAF with specific kwargs"""
    flow = MaskedAutoregressiveFlow(2, 2, 2, 2, **kwargs)
    x = torch.randn(10, 2)
    z, _ = flow.forward(x)
    assert z.shape == (10, 2)


@pytest.mark.parametrize("FlowClass", [RealNVP, NeuralSplineFlow])
def test_1d_inputs(FlowClass):
    """Assert an error is raised if 1-d inputs are specified."""
    with pytest.raises(ValueError) as excinfo:
        FlowClass(1, 2, 2, 2)
    assert "requires at least 2 dimensions" in str(excinfo.value)



# -*- coding: utf-8 -*-
"""
Test functions related to drawing posterior samples
"""
import logging
import numpy as np
import pytest
from unittest.mock import patch

from nessai.livepoint import numpy_array_to_live_points
from nessai.posterior import compute_weights, draw_posterior_samples
from nessai.utils.testing import assert_structured_arrays_equal

resampling_options = [
    "rejection_sampling",
    "importance_sampling",
    "multinomial_resampling",
]


@pytest.fixture()
def ns():
    return numpy_array_to_live_points(np.random.randn(20, 1), ["x"])


@pytest.fixture(params=resampling_options)
def method(request):
    return request.param


@pytest.mark.parametrize("nlive", [10, 10 * np.ones(20)])
@pytest.mark.parametrize("expectation", ["logt", "t"])
def test_compute_weights(nlive, expectation):
    """Test computing the weights for set of likelihood values."""
    log_l = np.random.randn(20)

    log_z, log_w = compute_weights(log_l, nlive, expectation=expectation)

    assert len(log_w) == len(log_l)
    assert np.isfinite(log_z)


@pytest.mark.parametrize("expectation", ["logt", "t"])
def test_compute_weights_correct_weights(expectation):
    """Assert the weights (X_i) are correct"""
    nlive = 10
    log_l = np.random.randn(20)
    out = -np.log1p(1 / nlive) * np.ones(len(log_l))
    with patch("numpy.log1p", return_value=out) as mock_log1p, patch(
        "nessai.posterior.logsubexp", return_value=np.random.rand(21)
    ), patch("nessai.posterior.log_integrate_log_trap", return_value=0.0):
        compute_weights(log_l, nlive=nlive, expectation=expectation)
    if expectation == "t":
        mock_log1p.assert_called_once()
    else:
        mock_log1p.assert_not_called()


def test_compute_weights_invalid_nlive():
    """Assert an error is raised if nlive does not match the logs-likelihood"""
    with pytest.raises(
        ValueError, match=r"nlive and samples are different lengths"
    ):
        compute_weights([1, 2, 3], [4, 5])


def test_compute_weights_invalid_expectation():
    """Assert an error is raised if expectation is invalid"""
    with pytest.raises(
        ValueError, match=r"Expectation must be t or logt, got: a"
    ):
        compute_weights(np.random.randn(10), 10, expectation="a")


def test_draw_posterior_samples(ns, method):
    """Test drawing posterior samples."""
    ns["logL"] = np.log(np.random.rand(ns.size))
    ns["logP"] = np.zeros(ns.size)
    p = draw_posterior_samples(ns, nlive=10, method=method)
    assert np.isin(p, ns).all()


def test_draw_posterior_samples_w_weights(ns, method):
    """Test drawing samples when weights are specified"""
    log_w = np.log(np.random.rand(len(ns)))
    p = draw_posterior_samples(ns, log_w=log_w, method=method)
    assert len(p) > 0


def test_draw_posterior_samples_with_n(caplog, ns, method):
    """Test drawing samples with n specified"""
    n = 10
    log_w = np.log(np.random.rand(len(ns)))
    with caplog.at_level(logging.WARNING):
        post = draw_posterior_samples(ns, log_w=log_w, method=method, n=n)
    if method == "rejection_sampling":
        assert "Number of samples cannot be specified" in caplog.text
    else:
        assert len(post) == n


def test_draw_posterior_samples_indices(ns, method):
    """Assert the indices are returned when return_indices=True"""
    log_w = np.log(np.random.rand(len(ns)))
    post, indices = draw_posterior_samples(
        ns,
        log_w=log_w,
        method=method,
        return_indices=True,
    )
    assert_structured_arrays_equal(post, ns[indices])


def test_draw_posterior_unknown_method(ns):
    """Assert an error is raised if the method is not known"""
    with pytest.raises(ValueError) as excinfo:
        draw_posterior_samples(ns, nlive=10, method="not_a_method")
    assert "Unknown method of drawing posterior samples: not_a_method" in str(
        excinfo.value
    )


@pytest.mark.slow_integration_test
def test_compute_weights_vs_log_posterior_weights(model, tmp_path):
    """Test the two different methods for computing posterior weights and
    assert they return the sames value.

    Checks both the values of the weights and log-evidence.
    """
    from nessai.flowsampler import FlowSampler

    output = tmp_path / "posterior_comparison"
    output.mkdir()
    fs = FlowSampler(
        model, output=output, nlive=100, checkpointing=False, plot=False
    )
    fs.run(save=False, plot=False)

    log_z_0, log_w_0 = compute_weights(fs.nested_samples["logL"], fs.ns.nlive)
    log_w_1 = fs.ns.state.log_posterior_weights

    np.testing.assert_array_equal(log_w_1, log_w_0)
    assert log_z_0 == fs.ns.state.log_evidence



# -*- coding: utf-8 -*-
"""
Test the combined reparameterisation class.
"""
import numpy as np
import pytest
from unittest.mock import MagicMock, call, create_autospec, patch

from nessai.reparameterisations import (
    Angle,
    CombinedReparameterisation,
    Reparameterisation,
    RescaleToBounds,
)
from nessai.livepoint import empty_structured_array
from nessai.utils.testing import assert_structured_arrays_equal


@pytest.fixture
def reparam():
    # Use spec_set to raised an error if an unknown attribute is set
    # Use an instance so attributes from __init__ are included.
    return create_autospec(
        CombinedReparameterisation(), spec_set=True, instance=True
    )


def test_init_none():
    c = CombinedReparameterisation()
    assert c.parameters == []


def test_init_w_reparam():
    c = CombinedReparameterisation(RescaleToBounds("x", [0, 1]))
    assert c.parameters == ["x"]
    assert c.prime_parameters == ["x_prime"]


@pytest.mark.parametrize("reverse_order", [False, True])
def test_to_prime_order(reverse_order, reparam):
    """Assert order is correct depending on the value of the reversed flag"""
    order = [1, 2, 3]
    reparam.order = order
    reparam.reverse_order = reverse_order

    expected = [1, 2, 3]
    if reverse_order:
        expected = list(reversed(expected))

    out = CombinedReparameterisation.to_prime_order.__get__(reparam)
    assert list(out) == expected


@pytest.mark.parametrize("reverse_order", [False, True])
def test_from_prime_order(reverse_order, reparam):
    """Assert order is correct depending on the value of the reversed flag"""
    order = [1, 2, 3]
    reparam.order = order
    reparam.reverse_order = reverse_order

    expected = [3, 2, 1]
    if reverse_order:
        expected = list(reversed(expected))

    out = CombinedReparameterisation.from_prime_order.__get__(reparam)
    assert list(out) == expected


def test_add_single_reparameterisations():
    """Test the core functionality of adding reparameterisations"""
    r = RescaleToBounds(parameters="x", prior_bounds=[0, 1])
    c = CombinedReparameterisation()
    c.add_reparameterisations(r)
    assert c.parameters == ["x"]
    assert c.has_prime_prior is False


def test_add_multiple_reparameterisations(model):
    """
    Test adding multiple reparameterisations and using the reparameterisation.
    """
    r = [
        RescaleToBounds(
            parameters="x", prior_bounds=model.bounds["x"], prior="uniform"
        ),
        RescaleToBounds(
            parameters="y", prior_bounds=model.bounds["y"], prior="uniform"
        ),
    ]
    reparam = CombinedReparameterisation()
    reparam.add_reparameterisations(r)

    assert reparam.parameters == ["x", "y"]
    assert reparam.has_prime_prior is True

    n = 100
    x = model.new_point(N=n)
    x_prime = empty_structured_array(n, names=reparam.prime_parameters)
    log_j = 0

    x_re, x_prime_re, log_j_re = reparam.reparameterise(x, x_prime, log_j)

    assert reparam.x_prime_log_prior(x_prime_re) is not None

    assert_structured_arrays_equal(x, x_re)

    x_in = empty_structured_array(n, names=reparam.parameters)

    x_inv, x_prime_inv, log_j_inv = reparam.inverse_reparameterise(
        x_in, x_prime_re, log_j
    )

    assert_structured_arrays_equal(x, x_inv)
    assert_structured_arrays_equal(x_prime_re, x_prime_inv)
    np.testing.assert_array_equal(log_j_re, -log_j_inv)


def test_base_add_reparameterisation_missing_requirements(reparam):
    """Assert an error is raised if a required parameter is missing"""
    r = MagicMock(spec=Reparameterisation)
    r.requires = ["x"]

    reparam.parameters = ["y"]
    reparam.prime_parameters = ["y_prime"]

    with pytest.raises(RuntimeError) as excinfo:
        CombinedReparameterisation._add_reparameterisation(reparam, r)
    assert "missing requirement(s)" in str(excinfo.value)


def test_add_reparameterisation(reparam):
    """Assert add_reparameterisation calls the correct method.

    Should just call CombinedReparameterisation.add_reparameteristions
    """
    new_reparam = MagicMock(spec=Reparameterisation)
    CombinedReparameterisation.add_reparameterisation(reparam, new_reparam)
    reparam.add_reparameterisations.assert_called_once_with(new_reparam)


def test_add_reparameterisations_single(reparam):
    """Assert add_reparameterisations works correctly for a single reparam.

    Should convert to a list and then pass to `_add_reparameterisation`
    """
    reparam.parameters = []
    new_reparam = MagicMock(spec=Reparameterisation)
    with patch(
        "nessai.reparameterisations.combined.sort_reparameterisations",
        return_value=[new_reparam],
    ) as mock:
        CombinedReparameterisation.add_reparameterisations(
            reparam, new_reparam
        )
    reparam._add_reparameterisation.assert_called_once_with(new_reparam)
    mock.assert_called_once_with([new_reparam], existing_parameters=[])


def test_add_reparameterisations_multiple(reparam):
    """Assert add_reparameterisations works correctly with multiple reparams.

    Should call `_add_reparmeterisation` with each reparam.
    """
    parameters = ["x"]
    reparam.parameters = parameters
    r1 = MagicMock(spec=Reparameterisation)
    r2 = MagicMock(spec=Reparameterisation)
    reparams = [r1, r2]
    with patch(
        "nessai.reparameterisations.combined.sort_reparameterisations",
        return_value=[r1, r2],
    ) as mock:
        CombinedReparameterisation.add_reparameterisations(reparam, reparams)
    reparam._add_reparameterisation.assert_has_calls([call(r1), call(r2)])
    mock.assert_called_once_with(reparams, existing_parameters=parameters)


def test_check_order_valid(reparam, LightReparam):
    """Assert no error is raised if the order is valid"""
    d = dict(
        a=LightReparam("a", parameters=["x"], requires=[]),
        b=LightReparam("b", parameters=["y"], requires=["x"]),
        c=LightReparam("c", parameters=["z"], requires=["y"]),
    )
    reparam.__getitem__.side_effect = d.__getitem__
    reparam.from_prime_order = ["a", "b", "c"]

    CombinedReparameterisation.check_order(reparam)


def test_check_order_invalid(reparam, LightReparam):
    """Assert an error is raised if the order is invalid"""
    d = dict(
        a=LightReparam("a", parameters=["x"], requires=["z"]),
        b=LightReparam("b", parameters=["y"], requires=[]),
        c=LightReparam("c", parameters=["z"], requires=["y"]),
    )
    reparam.__getitem__.side_effect = d.__getitem__
    reparam.from_prime_order = ["a", "b", "c"]
    print(reparam["a"])

    with pytest.raises(
        RuntimeError, match="Order of reparameterisations is invalid *."
    ):
        CombinedReparameterisation.check_order(reparam)


@pytest.mark.parametrize("has", [False, True])
def test_has_prime_prior(reparam, has):
    """Assert correct value is returned for has_prime_prior.

    If any of the reparam are missing the prime prior should be False. If all
    of them have it, should be True.
    """
    r1 = MagicMock(spec=Reparameterisation)
    r2 = MagicMock(spec=Reparameterisation)
    r1.has_prime_prior = True
    r2.has_prime_prior = has
    reparam.values = MagicMock(return_value=[r1, r2])

    out = CombinedReparameterisation.has_prime_prior.__get__(reparam)
    assert out is has


@pytest.mark.parametrize("require", [False, True])
def test_requires_prime_prior(reparam, require):
    """Assert correct value is returned for requires prime prior.

    Should be True if any reparam requires the prime prior and False only if
    none of them do.
    """
    r1 = MagicMock(spec=Reparameterisation)
    r2 = MagicMock(spec=Reparameterisation)
    r1.requires_prime_prior = False
    r2.requires_prime_prior = require
    reparam.values = MagicMock(return_value=[r1, r2])

    out = CombinedReparameterisation.requires_prime_prior.__get__(reparam)
    assert out is require


def test_update_bounds(reparam):
    """Assert update bounds calls the method for each reparam"""
    x = [1, 2]
    # Angle doesn't have update bounds
    r1 = MagicMock(spec=Angle)
    # RescaleToBounds does have update bounds
    r2 = MagicMock(spec=RescaleToBounds)
    reparam.values = MagicMock(return_value=[r1, r2])

    assert not hasattr(r1, "update_bounds")

    CombinedReparameterisation.update_bounds(reparam, x)

    r2.update_bounds.assert_called_once_with(x)


def test_reset_inversion(reparam):
    """Assert reset inversion calls the method for each reparam"""
    # Angle doesn't have reset inversion
    r1 = MagicMock(spec=Angle)
    # RescaleToBounds does have reset inversion
    r2 = MagicMock(spec=RescaleToBounds)
    reparam.values = MagicMock(return_value=[r1, r2])

    assert not hasattr(r1, "reset_inversion")

    CombinedReparameterisation.reset_inversion(reparam)

    r2.reset_inversion.assert_called_once()


def test_update(reparam):
    """Assert update calls the update method for all reparameterisations."""
    r1 = MagicMock(spec=Angle)
    r2 = MagicMock(spec=RescaleToBounds)
    reparam.values = MagicMock(return_value=[r1, r2])

    x = np.array((1, 2), dtype=[("x", "f8"), ("y", "f8")])

    CombinedReparameterisation.update(reparam, x)

    r1.update.assert_called_once_with(x)
    r2.update.assert_called_once_with(x)


def test_log_prior(reparam):
    """Assert log_prior is only called for reparams with has_prior==True"""
    r1 = MagicMock(spec=Reparameterisation)
    r2 = MagicMock(spec=Reparameterisation)
    r3 = MagicMock(spec=Reparameterisation)
    r1.has_prior = False
    r2.has_prior = True
    r3.has_prior = True
    r1.log_prior = MagicMock()
    r2.log_prior = MagicMock(return_value=-5)
    r3.log_prior = MagicMock(return_value=-6)
    reparam.values = MagicMock(return_value=[r1, r2, r3])

    x = np.array([(1, 2)], dtype=[("x", "f8"), ("y", "f8")])

    out = CombinedReparameterisation.log_prior(reparam, x)

    assert out == -11
    r1.log_prior.assert_not_called()
    r2.log_prior.assert_called_once_with(x)
    r3.log_prior.assert_called_once_with(x)


def test_x_prime_log_prior(reparam):
    """Assert x_prime_log_prior is called for all reparams"""
    r1 = MagicMock(spec=Reparameterisation)
    r2 = MagicMock(spec=Reparameterisation)
    r1.x_prime_log_prior = MagicMock(return_value=-4)
    r2.x_prime_log_prior = MagicMock(return_value=-5)
    reparam.values = MagicMock(return_value=[r1, r2])

    x = np.array([(1, 2)], dtype=[("x", "f8"), ("y", "f8")])

    out = CombinedReparameterisation.x_prime_log_prior(reparam, x)

    r1.x_prime_log_prior.assert_called_once_with(x)
    r2.x_prime_log_prior.assert_called_once_with(x)
    assert out == -9



# -*- coding: utf-8 -*-
"""Basic tests for all of the included flows"""
import pytest
import torch
import torch.nn.functional as F
import numpy as np

from nessai.flows import RealNVP, MaskedAutoregressiveFlow, NeuralSplineFlow

flows = [RealNVP, MaskedAutoregressiveFlow, NeuralSplineFlow]


@pytest.fixture(params=[2, 4])
def data_dim(request):
    return request.param


@pytest.fixture()
def conditional_features():
    return 1


@pytest.fixture()
def n():
    return 1000


@pytest.fixture()
def x(n, data_dim):
    return torch.randn(n, data_dim).float()


@pytest.fixture()
def z(n, data_dim):
    return torch.randn(n, data_dim).float()


@pytest.fixture(params=flows)
def flow(request, data_dim):
    return request.param(data_dim, 8, 2, 2).eval()


@pytest.fixture(params=flows)
def conditional_flow(request, data_dim, conditional_features):
    return request.param(
        data_dim, 8, 2, 2, context_features=conditional_features
    ).eval()


@pytest.fixture(params=flows)
def flow_class(request):
    return request.param


@pytest.mark.parametrize(
    "kwargs",
    [
        dict(batch_norm_between_layers=True),
        dict(batch_norm_within_layers=True),
        dict(activation=F.relu),
        dict(dropout_probability=0.5),
    ],
)
def test_init(flow_class, kwargs):
    """Test init method with common kwargs"""
    flow = flow_class(2, 2, 2, 2, **kwargs)
    x = torch.randn(10, 2)
    z, _ = flow.forward(x)
    assert z.shape == (10, 2)


def test_forward(flow, x, n, data_dim):
    """Test the forward pass of the flow"""
    with torch.inference_mode():
        z, _ = flow.forward(x)
    assert z.shape == (n, data_dim)


def test_inverse(flow, z, n, data_dim):
    """Test the inverse method"""
    with torch.inference_mode():
        x, _ = flow.inverse(z)
    assert x.shape == (n, data_dim)


def test_sample(flow, n, data_dim):
    """Test the sample method"""
    x = flow.sample(n)
    assert x.shape == (n, data_dim)


def test_log_prob(flow, x, n):
    """Test the log prob method"""
    with torch.inference_mode():
        log_prob = flow.log_prob(x)
    assert log_prob.shape == (n,)


def test_base_distribution_log_prob(flow, z, n):
    """Test the bast distribution"""
    with torch.inference_mode():
        log_prob = flow.base_distribution_log_prob(z)
    assert log_prob.shape == (n,)


def test_forward_and_log_prob(flow, x, n, data_dim):
    """
    Test the forward and log prob method.

    Tests to ensure method runs and that it agrees with using forward
    and log_prob separately
    """
    with torch.inference_mode():
        z, log_prob = flow.forward_and_log_prob(x)
        z_target, _ = flow.forward(x)
        log_prob_target = flow.log_prob(x)
    np.testing.assert_array_equal(z.numpy(), z_target.numpy())
    np.testing.assert_array_equal(log_prob.numpy(), log_prob_target.numpy())


@pytest.mark.flaky(reruns=5)
def test_sample_and_log_prob(flow, n, data_dim):
    """
    Assert that samples are drawn with correct shape and that the log
    prob is correct.
    """
    with torch.inference_mode():
        x, log_prob = flow.sample_and_log_prob(n)
        log_prob_target = flow.log_prob(x)
    assert x.shape == (n, data_dim)
    np.testing.assert_array_almost_equal(
        log_prob.numpy(), log_prob_target.numpy(), decimal=5
    )


@pytest.mark.flaky(reruns=5)
def test_invertibility(flow, x):
    """Test to ensure flows are invertible"""
    with torch.inference_mode():
        z, log_J = flow.forward(x)
        x_out, log_J_out = flow.inverse(z)

    np.testing.assert_array_almost_equal(x.numpy(), x_out.numpy(), decimal=5)
    np.testing.assert_array_almost_equal(
        log_J.numpy(), -log_J_out.numpy(), decimal=5
    )


@pytest.mark.flaky(reruns=5)
@pytest.mark.integration_test
def test_sample_and_log_prob_conditional(
    conditional_flow, n, data_dim, conditional_features
):
    """Test method for conditional flows."""
    c = torch.randn(n, conditional_features)
    with torch.inference_mode():
        x, log_prob = conditional_flow.sample_and_log_prob(n, context=c)
        log_prob_target = conditional_flow.log_prob(x, context=c)
    assert x.shape == (n, data_dim)
    np.testing.assert_array_almost_equal(
        log_prob.numpy(), log_prob_target.numpy(), decimal=5
    )



# -*- coding: utf-8 -*-
"""
Object and functions to handle training the normalising flow.
"""
import copy
import logging
import numpy as np
import os
import shutil
import torch
from torch.nn.utils import clip_grad_norm_
from typing import Optional, Tuple

from .utils import update_config

from ..flows import configure_model, reset_weights, reset_permutations
from ..flows.base import BaseFlow
from ..plot import plot_loss
from ..utils import save_to_json, compute_minimum_distances

logger = logging.getLogger(__name__)


class FlowModel:
    """
    Object that contains the normalising flows and handles training and data
    pre-processing.

    Does NOT use structured arrays for live points, \
            :obj:`~nessai.proposal.base.Proposal`
    object should act as the interface between structured used by the sampler
    and unstructured arrays of live points used for training.

    Parameters
    ----------
    config : dict, optional
        Configuration used for the normalising flow. If None, default values
        are used.
    output : str, optional
        Path for output, this includes weights files and the loss plot. If
        not specified, the current working directory is used.
    """

    model_config = None
    noise_scale = None
    noise_type = None
    model: BaseFlow = None

    def __init__(self, config=None, output=None):
        if output is None:
            output = os.getcwd()
        self.model = None
        self.initialised = False
        self.output = output
        os.makedirs(self.output, exist_ok=True)
        self.setup_from_input_dict(config)
        self.weights_file = None

        self.device = None
        self.inference_device = None
        self.use_dataloader = False
        self._batch_size = None

    def save_input(self, config, output_file=None):
        """
        Save the dictionary used as an inputs as a JSON file in the output
        directory.

        Parameters
        ----------
        config : dict
            Dictionary to save.
        output_file : str, optional
            File to save the config to.
        """
        config = copy.deepcopy(config)
        if output_file is None:
            output_file = os.path.join(self.output, "flow_config.json")
        for k, v in list(config.items()):
            if isinstance(v, np.ndarray):
                config[k] = np.array2string(config[k], separator=",")
        for k, v in list(config["model_config"].items()):
            if isinstance(v, np.ndarray):
                config["model_config"][k] = np.array2string(
                    config["model_config"][k], separator=","
                )

        if "flow" in config["model_config"]:
            config["model_config"]["flow"] = str(
                config["model_config"]["flow"]
            )

        save_to_json(config, output_file)

    def setup_from_input_dict(self, config):
        """
        Setup the trainer from a dictionary, all keys in the dictionary are
        added as methods to the object. Input is automatically saved.

        Parameters
        ----------
        config : dict
            Dictionary with parameters that are used to update the defaults.
        """
        config = update_config(config)
        logger.debug(f"Flow configuration: {config}")
        for key, value in config.items():
            setattr(self, key, value)
        self.save_input(config)

    def update_mask(self):
        """Method to update the ask upon calling ``initialise``

        By default the mask is left unchanged.
        """
        pass

    def get_optimiser(self, optimiser="adam", **kwargs):
        """
        Get the optimiser and ensure it is always correctly initialised.

        Returns
        -------
        :obj:`torch.optim.Adam`
            Instance of the Adam optimiser from torch.optim
        """
        optimisers = {
            "adam": (torch.optim.Adam, {"weight_decay": 1e-6}),
            "adamw": (torch.optim.AdamW, {}),
            "sgd": (torch.optim.SGD, {}),
        }
        if self.model is None:
            raise RuntimeError("Cannot initialise optimiser before model")
        optim, default_kwargs = optimisers.get(optimiser.lower())
        default_kwargs.update(kwargs)
        return optim(self.model.parameters(), lr=self.lr, **default_kwargs)

    def initialise(self):
        """
        Initialise the model and optimiser.

        This includes:

            - Updating the model configuration
            - Initialising the normalising flow
            - Initialising the optimiser
            - Configuring the inference device
        """
        self.update_mask()
        self.model, self.device = configure_model(self.model_config)
        logger.debug(f"Training device: {self.device}")
        self.inference_device = torch.device(
            self.model_config.get("inference_device_tag", self.device)
            or self.device
        )
        logger.debug(f"Inference device: {self.inference_device}")

        self._optimiser = self.get_optimiser(
            self.optimiser, **self.optimiser_kwargs
        )
        self.initialised = True

    def move_to(self, device, update_default=False):
        """Move the flow to a different device.

        Parameters
        ----------
        device : str
            Device to move flow to.
        update_default : bool, optional
            If True, the default device for the flow (and data) will be
            updated.
        """
        device = torch.device(device)
        self.model.to(device)
        if update_default:
            self.device = device

    @staticmethod
    def check_batch_size(x, batch_size, min_fraction=0.1):
        """Check that the batch size is valid.

        Tries to ensure that the last batch is at least a minimum fraction of
        the size of the batch size.

        Parameters
        ----------
        x : numpy.ndarray
            Training data.
        batch_size : int
            The user-specified batch size
        min_fraction : float
            Fraction of user-specified batch size to use a lower limit.
        """
        logger.debug("Checking batch size")
        if batch_size == 1:
            raise ValueError("Cannot use a batch size of 1!")
        min_batch_size = int(min_fraction * batch_size)
        final_batch_size = len(x) % batch_size
        if final_batch_size and (final_batch_size < min_batch_size):
            logger.debug(
                "Adjusting batch size to ensure final batch has at least "
                f"{min_batch_size} samples in it."
            )
            while True:
                batch_size -= 1
                final_batch_size = len(x) % batch_size
                if batch_size < 2:
                    raise RuntimeError("Could not find a valid batch size")
                elif (final_batch_size == 0) or (
                    final_batch_size >= min_batch_size
                ):
                    break
                elif (batch_size <= min_batch_size) and final_batch_size > 1:
                    logger.warning(
                        f"Batch size is less than {min_batch_size} but valid. "
                        f"Setting batch size to: {batch_size}"
                    )
                    break
        logger.debug(f"Using valid batch size of: {batch_size}")
        return batch_size

    def prep_data(
        self,
        samples,
        val_size,
        batch_size,
        weights=None,
        use_dataloader=False,
        conditional=None,
    ):
        """
        Prep data and return dataloaders for training

        Parameters
        ----------
        samples : array_like
            Array of samples to split in to training and validation.
        val_size : float
            Float between 0 and 1 that defines the fraction of data used for
            validation.
        batch_size : int
            Batch size used when constructing dataloaders.
        use_dataloader : bool, optional
            If True data is returned in a dataloader else a tensor is returned.
        weights : array_like, optional
            Array of weights for each sample, weights will used when computing
            the loss.

        Returns
        -------
        train_data, val_data :
            Training and validation data as either a tensor or dataloader
        """
        if not self.initialised:
            self.initialise()

        if not np.isfinite(samples).all():
            raise ValueError("Cannot train with non-finite samples!")

        if weights is not None and conditional is not None:
            raise RuntimeError("weights and conditional inputs not supported")

        idx = np.random.permutation(samples.shape[0])
        samples = samples[idx]
        if weights is not None:
            if not np.isfinite(weights).all():
                raise ValueError("Weights contain non-finite values!")
            weights = weights[idx]
            use_dataloader = True
        if conditional is not None:
            conditional = conditional[idx]
            use_dataloader = True

        logger.debug("N input samples: {}".format(len(samples)))

        # setup data loading
        if val_size is None:
            val_size = 0
        n = int((1 - val_size) * samples.shape[0])
        x_train, x_val = samples[:n], samples[n:]
        if weights is not None:
            weights_train = weights[:n]
            weights_val = weights[n:]
        if conditional is not None:
            conditional_train = conditional[:n]
            conditional_val = conditional[n:]
        logger.debug(f"{x_train.shape} training samples")
        logger.debug(f"{x_val.shape} validation samples")

        if isinstance(batch_size, bool) or not isinstance(batch_size, int):
            if batch_size == "all" or batch_size is None:
                batch_size = x_train.shape[0]
            else:
                raise RuntimeError(f"Unknown batch size: {batch_size}")

        batch_size = self.check_batch_size(x_train, batch_size)
        self._batch_size = batch_size
        # Validation size cannot be bigger than the batch size
        val_batch_size = min(len(x_val), batch_size) if len(x_val) else None

        dtype = torch.get_default_dtype()
        logger.debug(f"Using dtype {dtype} for tensors")
        if use_dataloader:
            logger.debug("Using dataloaders")
            train_tensors = [torch.from_numpy(x_train).type(dtype)]
            val_tensors = [torch.from_numpy(x_val).type(dtype)]

            if weights is not None:
                train_tensors.append(
                    torch.from_numpy(weights_train).type(dtype)
                )
                val_tensors.append(torch.from_numpy(weights_val).type(dtype))

            if conditional is not None:
                train_tensors.append(
                    torch.from_numpy(conditional_train).type(dtype)
                )
                val_tensors.append(
                    torch.from_numpy(conditional_val).type(dtype)
                )

            train_dataset = torch.utils.data.TensorDataset(*train_tensors)
            train_data = torch.utils.data.DataLoader(
                train_dataset, batch_size=batch_size, shuffle=True
            )

            val_dataset = torch.utils.data.TensorDataset(*val_tensors)
            val_data = torch.utils.data.DataLoader(
                val_dataset, batch_size=val_batch_size, shuffle=False
            )
        else:
            logger.debug("Using tensors")
            train_data = torch.from_numpy(x_train).type(dtype).to(self.device)
            val_data = torch.from_numpy(x_val).type(dtype).to(self.device)

        return train_data, val_data, batch_size

    def end_iteration(self):
        """Calls any functions that should be applied at the end of the \
            iteration.

        This functions is called after the flow has been updated on all batches
        of data but before the validation step.

        Calls :py:meth:`nessai.flows.base.BaseFlow.end_iteration`
        """
        self.model.end_iteration()

    def _train(
        self,
        train_data,
        noise_scale=0.0,
        is_dataloader=False,
        weighted=False,
        is_conditional=False,
    ):
        """
        Loop over the data and update the weights

        Parameters
        ----------
        train_data : :obj:`torch.util.data.Dataloader` or :obj:`torch.Tensor`
            Training data. If a tensor is provided, it is split into batches
            using the batch size.
        noise_scale : float, optional
            Scale of Gaussian noise added to data.
        is_dataloader : bool, optional
            Must be True when using a dataloader
        weighted : bool
            If True the weighted KL will be used to compute the loss. Requires
            data to include weights.
        is_conditional : bool
            If True, the conditional argument will be passed to the flow.

        Returns
        -------
        float
            Mean of training loss for each batch.
        """
        model = self.model
        model.train()
        train_loss = 0

        if hasattr(model, "loss_function"):
            loss_fn = model.loss_function
        elif weighted:

            def loss_fn(data, weights):
                return -torch.sum(model.log_prob(data) * weights) / torch.sum(
                    weights
                )

        else:

            def loss_fn(data, conditional):
                return -model.log_prob(data, conditional).mean()

        if not is_dataloader:
            p = torch.randperm(train_data.shape[0])
            train_data = train_data[p, :].split(self._batch_size)

        n = 0
        for data in train_data:
            if is_dataloader:
                x = data[0].to(self.device)
            else:
                x = data
            if noise_scale:
                x = x + noise_scale * torch.randn_like(x)
            for param in model.parameters():
                param.grad = None
            if weighted:
                weights = data[1].to(self.device)
                loss = loss_fn(x, weights)
            else:
                if is_conditional:
                    conditional = data[1].to(self.device)
                else:
                    conditional = None
                loss = loss_fn(x, conditional)
            train_loss += loss.item()
            loss.backward()
            if self.clip_grad_norm:
                clip_grad_norm_(model.parameters(), self.clip_grad_norm)
            self._optimiser.step()
            n += 1

        self.end_iteration()

        if self.annealing:
            self.scheduler.step()

        return train_loss / n

    def _validate(
        self,
        val_data,
        is_dataloader=False,
        weighted=False,
        is_conditional=False,
    ):
        """
        Loop over the data and get validation loss

        Parameters
        ----------
        val_data : :obj:`torch.util.data.Dataloader` or :obj:`torch.Tensor
            Dataloader with data to validate on
        is_dataloader : bool, optional
            Boolean to indicate if the data is a dataloader
        weighted : bool
            If True the weighted KL will be used to compute the loss. Requires
            data to include weights.
        is_conditional : bool
            If True, the conditional argument will be passed to the flow.

        Returns
        -------
        float
            Mean of training loss for each batch.
        """
        if not len(val_data):
            return np.nan
        model = self.model
        model.eval()
        val_loss = 0

        if hasattr(model, "loss_function"):
            loss_fn = model.loss_function
        elif weighted:

            def loss_fn(data, weights):
                return -torch.sum(model.log_prob(data) * weights) / torch.sum(
                    weights
                )

        else:

            def loss_fn(data, conditional):
                return -model.log_prob(data, conditional).mean()

        if is_dataloader:
            n = 0
            for data in val_data:
                if is_dataloader:
                    x = data[0].to(self.device)
                if weighted:
                    weights = data[1].to(self.device)
                    with torch.inference_mode():
                        val_loss += loss_fn(x, weights).item()
                else:
                    if is_conditional:
                        conditional = data[1].to(self.device)
                    else:
                        conditional = None
                    with torch.inference_mode():
                        val_loss += loss_fn(x, conditional).item()
                n += 1

            return val_loss / n
        else:
            with torch.inference_mode():
                val_loss += loss_fn(val_data, None).item()
            return val_loss

    def finalise(self):
        """Method to finalise the flow before using it for inference."""
        logger.debug("Finalising model before inference.")
        self.model.finalise()

    def train(
        self,
        samples,
        weights=None,
        conditional=None,
        max_epochs=None,
        patience=None,
        output=None,
        val_size=None,
        plot=True,
    ):
        """
        Train the flow on a set of samples.

        Allows for training parameters to specified instead of those
        given in initial config.

        Parameters
        ----------
        samples : ndarray
            Unstructured numpy array containing data to train on
        weights : array_like
            Array of weights to use with the weight KL when computing the loss.
        max_epochs : int, optional
            Maximum number of epochs that is used instead of value
            in the configuration.
        patience : int, optional
            Patience in number of epochs that is used instead of value
            in the configuration.
        val_size : float, optional
           Fraction of the samples to use for validation
        output : str, optional
            Path to output directory that is used instead of the path
            specified when the object is initialised
        plot : bool, optional
            Boolean to enable or disable plotting the loss function

        Returns
        -------
        history : dict
            Dictionary that contains the training and validation losses.
        """
        if not self.initialised:
            logger.debug("Initialising")
            self.initialise()

        if not np.isfinite(samples).all():
            raise ValueError("Training data is not finite")

        logger.debug("Data summary:")
        logger.debug(f"Mean: {np.mean(samples, axis=0)}")
        logger.debug(f"Standard deviation: {np.std(samples, axis=0, ddof=1)}")

        if output is None:
            output = self.output
        else:
            os.makedirs(output, exist_ok=True)

        if val_size is None:
            val_size = self.val_size

        if val_size == 0.0:
            validate = False
        else:
            validate = True

        if self.noise_type == "adaptive":
            noise_scale = self.noise_scale * np.mean(
                compute_minimum_distances(samples)
            )
            logger.debug(f"Using adaptive scale: {noise_scale:.3f}")
        elif self.noise_type == "constant":
            noise_scale = self.noise_scale
        else:
            logger.debug("No noise will be added in training")
            noise_scale = None

        self.move_to(self.device)

        weighted = True if weights is not None else False
        is_conditional = True if conditional is not None else False
        use_dataloader = self.use_dataloader or weighted or is_conditional

        train_data, val_data, _ = self.prep_data(
            samples,
            val_size=val_size,
            batch_size=self.batch_size,
            weights=weights,
            conditional=conditional,
            use_dataloader=use_dataloader,
        )

        if max_epochs is None:
            max_epochs = self.max_epochs
        if self.annealing:
            self.scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(
                self._optimiser, max_epochs
            )
        if patience is None:
            patience = self.patience
        best_epoch = 0
        best_val_loss = np.inf
        best_model = copy.deepcopy(self.model.state_dict())
        logger.debug("Starting training")
        logger.debug("Training parameters:")
        logger.debug(f"Max. epochs: {max_epochs}")
        logger.debug(f"Patience: {patience}")
        logger.debug(f"Training with {samples.shape[0]} samples")

        history = dict(loss=[], val_loss=[])

        current_weights_file = os.path.join(output, "model.pt")

        for epoch in range(1, max_epochs + 1):

            loss = self._train(
                train_data,
                noise_scale=noise_scale,
                is_dataloader=use_dataloader,
                weighted=weighted,
                is_conditional=is_conditional,
            )
            val_loss = self._validate(
                val_data,
                is_dataloader=use_dataloader,
                weighted=weighted,
                is_conditional=is_conditional,
            )
            history["loss"].append(loss)
            history["val_loss"].append(val_loss)

            if validate and (val_loss < best_val_loss):
                best_epoch = epoch
                best_val_loss = val_loss
                best_model = copy.deepcopy(self.model.state_dict())

            if not epoch % 50:
                logger.debug(
                    f"Epoch {epoch}: loss: {loss:.3} val loss: {val_loss:.3}"
                )

            if validate and (epoch - best_epoch > patience):
                logger.debug(f"Epoch {epoch}: Reached patience")
                break

        logger.debug("Finished training")

        # Make sure cache for LU is reset.
        self.model.train()
        self.model.eval()
        if validate:
            logger.debug(f"Loading best model from epoch {best_epoch}")
            self.model.load_state_dict(best_model)
        self.finalise()

        self.save_weights(current_weights_file)
        self.move_to(self.inference_device)
        self.model.eval()

        if plot:
            plot_loss(
                epoch, history, filename=os.path.join(output, "loss.png")
            )
        return history

    def save_weights(self, weights_file):
        """
        Save the weights file. If the file already exists move it to
        ``<weights_file>.old`` and then save the file.

        Parameters
        ----------
        weights_file : str
            Path to to file to save weights. Recommended file type is ``.pt``.
        """
        if os.path.exists(weights_file):
            shutil.move(weights_file, weights_file + ".old")

        torch.save(self.model.state_dict(), weights_file)
        self.weights_file = weights_file

    def load_weights(self, weights_file):
        """
        Load weights for the model and initialises the model if it is not
        initialised. The weights_file attribute is also updated.

        Model is loaded in evaluation mode (``model.eval()``)

        Parameters
        ----------
        weights_files : str
            Path to weights file
        """
        # TODO: these two methods are basically the same
        if not self.initialised:
            self.initialise()
        self.model.load_state_dict(torch.load(weights_file))
        self.model.eval()
        self.weights_file = weights_file

    def reload_weights(self, weights_file):
        """
        Tries to the load the weights file and if not, tries to load
        the weights file stored internally.

        Parameters
        ----------
        weights_files : str
            Path to weights file
        """
        if weights_file is None:
            weights_file = self.weights_file
        logger.debug(f"Reloading weights from {weights_file}")
        self.load_weights(weights_file)

    def reset_model(self, weights=True, permutations=False):
        """
        Reset the weights of the model and optimiser.

        Parameters
        ----------
        weights : bool, optional
            If true the model weights are reset.
        permutations : bool, optional
            If true any permutations (linear transforms) are reset.
        """
        if not any([weights, permutations]):
            logger.debug("Nothing to reset")
            return
        if weights and permutations:
            logger.debug("Complete reset of model")
            self.model, self.device = configure_model(self.model_config)
        elif weights:
            self.model.apply(reset_weights)
            logger.debug("Reset weights")
        elif permutations:
            self.model.apply(reset_permutations)
            logger.debug("Reset linear transforms")
        self._optimiser = self.get_optimiser(
            self.optimiser, **self.optimiser_kwargs
        )
        logger.debug("Resetting optimiser")

    def numpy_array_to_tensor(self, array: np.ndarray, /) -> torch.Tensor:
        """Convert a numpy array to a tensor and move it to the device"""
        return (
            torch.from_numpy(array)
            .type(torch.get_default_dtype())
            .to(self.model.device)
        )

    def forward_and_log_prob(
        self, x: np.ndarray, conditional: Optional[np.ndarray] = None
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Forward pass through the model and return the samples in the latent
        space with their log probabilities

        Parameters
        ----------
        x : ndarray
            Array of samples

        Returns
        -------
        z : ndarray
            Samples in the latent space
        log_prob : ndarray
            Log probabilities for each samples
        """
        x = self.numpy_array_to_tensor(x)
        if conditional is not None:
            conditional = self.numpy_array_to_tensor(conditional)
        self.model.eval()
        with torch.inference_mode():
            z, log_prob = self.model.forward_and_log_prob(
                x, context=conditional
            )
        z = z.detach().cpu().numpy().astype(np.float64)
        log_prob = log_prob.detach().cpu().numpy().astype(np.float64)
        return z, log_prob

    def log_prob(
        self, x: np.ndarray, conditional: Optional[np.ndarray] = None
    ) -> np.ndarray:
        """Compute the log-probability of a sample.

        Parameters
        ----------
        x : ndarray
            Array of samples in the X-prime space.

        Returns
        -------
        ndarray
            Array of log-probabilities.
        """
        x = self.numpy_array_to_tensor(x)
        if conditional is not None:
            conditional = self.numpy_array_to_tensor(conditional)
        self.model.eval()
        with torch.inference_mode():
            log_prob = self.model.log_prob(x, context=conditional)
        log_prob = log_prob.cpu().numpy().astype(np.float64)
        return log_prob

    def sample(
        self, n: int = 1, conditional: Optional[np.ndarray] = None
    ) -> np.ndarray:
        """Sample from the flow.

        Parameters
        ----------
        n : int
            Number of samples to draw
        conditional : numpy.ndarray
            Array of conditional inputs

        Returns
        -------
        numpy.ndarray
            Array of samples
        """
        if conditional is not None:
            conditional = self.numpy_array_to_tensor(conditional)
        with torch.inference_mode():
            x = self.model.sample(int(n), context=conditional)
        return x.cpu().numpy().astype(np.float64)

    def sample_latent_distribution(self, n: int = 1) -> np.ndarray:
        """Sample from the latent distribution.

        Parameters
        ----------
        n : int
            Number of samples to draw

        Returns
        -------
        numpy.ndarray
            Array of samples
        """
        with torch.inference_mode():
            z = self.model.sample_latent_distribution(n)
        return z.cpu().numpy().astype(np.float64)

    def sample_and_log_prob(
        self, N=1, z=None, alt_dist=None, conditional=None
    ):
        """
        Generate samples from samples drawn from the base distribution or
        and alternative distribution from provided latent samples

        Parameters
        ----------
        N : int, optional
            Number of samples to draw if z is not specified
        z : ndarray, optional
            Array of latent samples to map the the data space, if ``alt_dist``
            is not specified they are assumed to be drawn from the base
            distribution of the flow.
        alt_dist : :obj:`glasflow.nflows.distribution.Distribution`
            Distribution object from which the latent samples z were
            drawn from. Must have a ``log_prob`` method that accepts an
            instance of ``torch.Tensor``

        Returns
        -------
        samples : ndarray
            Array containing samples in the latent space.
        log_prob : ndarray
            Array containing the log probability that corresponds to each
            sample.
        """
        if self.model is None:
            raise RuntimeError("Model is not initialised yet!")
        if self.model.training:
            self.model.eval()
        if conditional is not None:
            conditional = self.numpy_array_to_tensor(conditional)
        if z is None:
            with torch.inference_mode():
                x, log_prob = self.model.sample_and_log_prob(
                    int(N), context=conditional
                )
        else:
            if alt_dist is not None:
                log_prob_fn = alt_dist.log_prob
            else:
                log_prob_fn = self.model.base_distribution_log_prob

            with torch.inference_mode():
                if isinstance(z, np.ndarray):
                    z = self.numpy_array_to_tensor(z)
                log_prob = log_prob_fn(z)
                x, log_J = self.model.inverse(z, context=conditional)
                log_prob -= log_J

        x = x.detach().cpu().numpy().astype(np.float64)
        log_prob = log_prob.detach().cpu().numpy().astype(np.float64)
        return x, log_prob

    def __getstate__(self):
        state = self.__dict__.copy()
        state["initialised"] = False
        del state["optimiser"]
        del state["model"]
        del state["model_config"]
        return state



# -*- coding: utf-8 -*-
"""
Various utilities for implementing normalising flows.
"""
import inspect
import logging
from typing import Optional, Type, Union
import warnings

from glasflow.nflows import transforms
from glasflow.nflows.distributions import Distribution
from glasflow.distributions import MultivariateUniform
import numpy as np
import torch
import torch.nn.functional as F

from .distributions import MultivariateNormal, ResampledGaussian
from .nets import MLP


logger = logging.getLogger(__name__)


def silu(x):
    """
    SiLU (Sigmoid-weighted Linear Unit) activation function.

    Also known as swish.

    Elfwing et al 2017: https://arxiv.org/abs/1702.03118v3
    """
    return torch.mul(x, torch.sigmoid(x))


def get_base_distribution(
    n_inputs: int, distribution: Union[str, Type[Distribution]], **kwargs
) -> Distribution:
    """Get the base distribution for a flow.

    Includes special configuration for certain distributions.

    Parameters
    ----------
    n_inputs : int
        Number of inputs to the distribution.
    distribution : Union[str, Type[glasflow.nflows.distribution.Distribution]]
        Distribution class or name of known distribution
    kwargs : Any
        Keyword arguments used when creating an instance of distribution.
    """
    distributions = {
        "mvn": MultivariateNormal,
        "normal": MultivariateNormal,
        "lars": ResampledGaussian,
        "resampled": ResampledGaussian,
        "uniform": MultivariateUniform,
    }

    DistClass = None

    if isinstance(distribution, str):
        DistClass = distributions.get(distribution.lower())
        if not DistClass:
            raise ValueError(f"Unknown distribution: {distribution}")
    elif inspect.isclass(distribution):
        logger.debug("Distribution is class. Creating an instance.")
        DistClass = distribution

    if DistClass:
        logger.debug("Creating instance of the base distribution")
        if DistClass is ResampledGaussian:
            n_layers = kwargs.pop("n_layers", 2)
            n_neurons = get_n_neurons(
                kwargs.pop("n_neurons", None), n_inputs=n_inputs
            )
            layers_list = n_layers * [n_neurons]
            logger.debug(
                f"LARS acceptance network will have {n_layers} layers with "
                f"{n_neurons} neurons each."
            )
            net_kwargs = kwargs.pop("net_kwargs", {})
            acc_fn = MLP(
                [n_inputs],
                [1],
                layers_list,
                activate_output=torch.sigmoid,
                **net_kwargs,
            )
            logger.debug(f"Other LARs kwargs: {kwargs}")
            dist = DistClass([n_inputs], acc_fn, **kwargs)
        elif DistClass is MultivariateUniform:
            dist = DistClass(
                low=torch.zeros(n_inputs, dtype=torch.get_default_dtype()),
                high=torch.ones(n_inputs, dtype=torch.get_default_dtype()),
            )
        else:
            dist = DistClass([n_inputs], **kwargs)
    elif distribution is None:
        dist = None
    else:
        dist = distribution
    return dist


def get_n_neurons(
    n_neurons: Optional[int] = None,
    n_inputs: Optional[int] = None,
    default: int = 8,
) -> int:
    """Get the number of neurons.

    Notes
    -----
    If :code:`n_inputs` is also specified then the options for
    :code:`n_neurons` are either a value that can be converted to an
    :code:`int` or one of the following:

        - :code:`'auto'` or :code:`'double'`: uses twice the number of inputs
        - :code:`'equal'`: uses the number of inputs
        - :code:`'half'`: uses half the number of inputs
        - :code:`None`: falls back to :code:`'auto'`

    Parameters
    ----------
    n_neurons : Optional[int]
        Number of neurons.
    n_inputs: Optional[int]
        Number of inputs.
    default : int
        Default value if :code:`n_neurons` and :code:`n_inputs` are not given.

    Returns
    -------
    int
        Number of neurons.

    Raises
    ------
    ValueError
        Raised if the number of inputs could not be converted to an integer.
    """
    if n_inputs is None:
        if n_neurons is None:
            n = default
        else:
            n = n_neurons
    else:
        if n_neurons is None or n_neurons in {"auto", "double"}:
            n = 2 * n_inputs
        elif n_neurons == "equal":
            n = n_inputs
        elif n_neurons == "half":
            n = n_inputs // 2
        else:
            n = n_neurons
    try:
        n = int(n)
    except ValueError:
        raise ValueError(
            "Could not get number of neurons. `n_neurons` was set to "
            f"`{n_neurons}` which could not be translated to a valid int. If "
            "using `auto`, `double`, `equal` or `half`, check that `n_inputs` "
            "is set."
        )
    return n


def configure_model(config):
    """
    Setup the flow form a configuration dictionary.
    """
    from .realnvp import RealNVP
    from .maf import MaskedAutoregressiveFlow
    from .nsf import NeuralSplineFlow
    from ..flowmodel import config as fmconfig

    kwargs = {}
    flows = {
        "realnvp": RealNVP,
        "maf": MaskedAutoregressiveFlow,
        "frealnvp": RealNVP,
        "spline": NeuralSplineFlow,
        "nsf": NeuralSplineFlow,
    }
    activations = {"relu": F.relu, "tanh": F.tanh, "swish": silu, "silu": silu}

    config = config.copy()

    if not isinstance(config["n_inputs"], int):
        raise TypeError("Number of inputs (n_inputs) must be an int")

    allowed_keys = set(fmconfig.DEFAULT_MODEL_CONFIG.keys())
    extra_keys = set(config.keys()) - allowed_keys
    if extra_keys:
        raise RuntimeError(
            f"Unknown keys in model config: {extra_keys}. "
            f"Known keys are: {allowed_keys}"
        )

    k = config.get("kwargs", None)
    if k is not None:
        if "activation" in k and isinstance(k["activation"], str):
            try:
                k["activation"] = activations[k["activation"]]
            except KeyError as e:
                raise RuntimeError(f"Unknown activation function: {e}")

        kwargs.update(k)

    dist_kwargs = config.pop("distribution_kwargs", None)
    if dist_kwargs is None:
        dist_kwargs = {}
    distribution = get_base_distribution(
        config["n_inputs"],
        config.pop("distribution", None),
        **dist_kwargs,
    )
    if distribution:
        kwargs["distribution"] = distribution

    fc = config.get("flow", None)
    ftype = config.get("ftype", None)
    if fc is not None:
        model = fc(
            config["n_inputs"],
            config["n_neurons"],
            config["n_blocks"],
            config["n_layers"],
            **kwargs,
        )
    elif ftype is not None:
        if ftype.lower() not in flows:
            raise RuntimeError(
                f"Unknown flow type: {ftype}. Choose from:" f"{flows.keys()}"
            )
        if (
            ("mask" in kwargs and kwargs["mask"] is not None)
            or ("net" in kwargs and kwargs["net"] is not None)
        ) and ftype.lower() not in ["realnvp", "frealnvp"]:
            raise RuntimeError(
                "Custom masks and networks are only " "supported for RealNVP"
            )

        model = flows[ftype.lower()](
            config["n_inputs"],
            config["n_neurons"],
            config["n_blocks"],
            config["n_layers"],
            **kwargs,
        )
    else:
        raise RuntimeError("Must specify either 'flow' or 'ftype'.")

    device = torch.device(config.get("device_tag", "cpu"))
    if device != "cpu":
        try:
            model.to(device)
        except RuntimeError as e:
            device = torch.device("cpu")
            logger.warning(
                "Could not send the normalising flow to the "
                f"specified device {config['device_tag']} send to CPU "
                f"instead. Error raised: {e}"
            )
    logger.debug("Flow model:")
    logger.debug(model)

    model.device = device

    return model, device


def reset_weights(module):
    """Reset parameters of a given module in place.

    Uses the ``reset_parameters`` method from ``torch.nn.Module``

    Also checks the following modules from glasflow.nflows

    - glasflow.nflows.transforms.normalization.BatchNorm

    Parameters
    ----------
    module : :obj:`torch.nn.Module`
        Module to reset
    """
    if hasattr(module, "reset_parameters"):
        module.reset_parameters()
    elif isinstance(module, transforms.BatchNorm):
        # glasflow.nflows BatchNorm does not have a weight reset, so must
        # be done manually
        constant = np.log(np.exp(1 - module.eps) - 1)
        module.unconstrained_weight.data.fill_(constant)
        module.bias.data.zero_()
        module.running_mean.zero_()
        module.running_var.fill_(1)
    else:
        logger.warning(f"Could not reset: {module}")


def reset_permutations(module):
    """Resets permutations and linear transforms for a given module in place.

    Resets using the original initialisation method. This needed since they
    do not have a ``reset_parameters`` method.

    Parameters
    ----------
    module : :obj:`torch.nn.Module`
        Module to reset
    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        from .transforms import LULinear
    if isinstance(module, (transforms.LULinear, LULinear)):
        module.cache.invalidate()
        module._initialize(identity_init=True)
    elif isinstance(module, transforms.RandomPermutation):
        module._permutation = torch.randperm(len(module._permutation))


def create_linear_transform(linear_transform, features):
    """Function for creating linear transforms.

    Parameters
    ----------
    linear_transform : {'permutation', 'lu', 'svd'}
        Linear transform to use.
    featres : int
        Number of features.
    """
    if linear_transform.lower() == "permutation":
        return transforms.RandomPermutation(features=features)
    elif linear_transform.lower() == "lu":
        return transforms.CompositeTransform(
            [
                transforms.RandomPermutation(features=features),
                transforms.LULinear(
                    features, identity_init=True, using_cache=True
                ),
            ]
        )
    elif linear_transform.lower() == "svd":
        return transforms.CompositeTransform(
            [
                transforms.RandomPermutation(features=features),
                transforms.SVDLinear(
                    features, num_householder=10, identity_init=True
                ),
            ]
        )
    else:
        raise ValueError(
            f"Unknown linear transform: {linear_transform}. "
            "Choose from: {permutation, lu, svd}."
        )


def create_pre_transform(pre_transform, features, **kwargs):
    """Create a pre transform.

    Parameters
    ----------
    pre_transform : str, {logit, batch_norm}
        Name of the transform
    features : int
        Number of input features
    kwargs :
        Keyword arguments passed to the transform class.
    """
    if pre_transform == "logit":
        return transforms.Logit(**kwargs)
    elif pre_transform == "batch_norm":
        return transforms.BatchNorm(features=features, **kwargs)
    else:
        raise ValueError(f"Unknown pre-transform: {pre_transform}")



#!/usr/bin/env python
"""
Example of running nessai with calibration uncertainty. Nessai can run with
calibration uncertainty but it has not been extensively tested.

This example should take ~30 minutes to run.

Adapted from the example included in bilby.
"""
import numpy as np
import bilby

# Standard configuration using bilby
duration = 4.0
sampling_frequency = 2048.0

outdir = "./outdir/"
label = "calibration_example"
bilby.core.utils.setup_logger(outdir=outdir, label=label)

np.random.seed(150914)

# Injection parameters for a GW150914-like BBH.
injection_parameters = dict(
    mass_1=36.0,
    mass_2=29.0,
    a_1=0.4,
    a_2=0.3,
    tilt_1=0.5,
    tilt_2=1.0,
    phi_12=1.7,
    phi_jl=0.3,
    luminosity_distance=2000.0,
    theta_jn=0.4,
    psi=2.659,
    phase=1.3,
    geocent_time=1126259642.413,
    ra=1.375,
    dec=-1.2108,
)

waveform_arguments = dict(
    waveform_approximant="IMRPhenomPv2", reference_frequency=50.0
)

# Create the waveform_generator
waveform_generator = bilby.gw.WaveformGenerator(
    duration=duration,
    sampling_frequency=sampling_frequency,
    frequency_domain_source_model=bilby.gw.source.lal_binary_black_hole,
    parameters=injection_parameters,
    waveform_arguments=waveform_arguments,
)

# Set up interferometers
ifos = bilby.gw.detector.InterferometerList(["H1", "L1"])
for ifo in ifos:
    injection_parameters.update(
        {
            "recalib_{}_amplitude_{}".format(ifo.name, ii): 0.05
            for ii in range(5)
        }
    )
    injection_parameters.update(
        {"recalib_{}_phase_{}".format(ifo.name, ii): 0.01 for ii in range(5)}
    )
    ifo.calibration_model = bilby.gw.calibration.CubicSpline(
        prefix="recalib_{}_".format(ifo.name),
        minimum_frequency=ifo.minimum_frequency,
        maximum_frequency=ifo.maximum_frequency,
        n_points=5,
    )
ifos.set_strain_data_from_power_spectral_densities(
    sampling_frequency=sampling_frequency, duration=duration
)
ifos.inject_signal(
    parameters=injection_parameters, waveform_generator=waveform_generator
)

# Set up prior
priors = bilby.gw.prior.BBHPriorDict()
priors["geocent_time"] = bilby.core.prior.Uniform(
    minimum=injection_parameters["geocent_time"] - 0.1,
    maximum=injection_parameters["geocent_time"] + 0.1,
    name="geocent_time",
    latex_label="$t_c$",
    unit="$s$",
)
fixed_parameters = [
    "a_1",
    "a_2",
    "tilt_1",
    "tilt_2",
    "phi_jl",
    "phi_12",
    "psi",
    "geocent_time",
    "luminosity_distance",
]

for key in injection_parameters:
    if "recalib" in key:
        priors[key] = injection_parameters[key]
    if key in fixed_parameters:
        priors[key] = injection_parameters[key]

calib_amplitudes = [
    "recalib_H1_amplitude_0",
    "recalib_L1_amplitude_0",
    "recalib_H1_amplitude_1",
    "recalib_L1_amplitude_1",
]

calib_phases = [
    "recalib_H1_phase_0",
    "recalib_L1_phase_0",
    "recalib_H1_phase_1",
    "recalib_L1_phase_1",
]

# Entry 8 is the detector (H or L)
for name in calib_amplitudes:
    priors[name] = bilby.prior.Gaussian(
        mu=0, sigma=0.2, name=name, latex_label=f"{name[8]}1 $A_{name[-1]}$"
    )

for name in calib_phases:
    priors[name] = bilby.prior.Gaussian(
        mu=0, sigma=0.1, name=name, latex_label=f"{name[8]}1 $P_{name[-1]}$"
    )

# Initialise the likelihood by passing in the interferometer data (IFOs) and
# the waveform generator
likelihood = bilby.gw.GravitationalWaveTransient(
    interferometers=ifos,
    waveform_generator=waveform_generator,
    priors=priors,
    phase_marginalization=True,
)

# Run sampler.
# By default the calibration parameters will not be reparameterised.
# Here we show how you could set a reparameterisation for these parameters.
result = bilby.run_sampler(
    resume=False,
    likelihood=likelihood,
    priors=priors,
    conversion_function=bilby.gw.conversion.generate_all_bbh_parameters,
    injection_parameters=injection_parameters,
    outdir=outdir,
    label=label,
    plot=True,
    sampler="nessai",
    flow_class="gwflowproposal",
    reparameterisations={"z-score": {"parameters": "recalib.*"}},
)



# -*- coding: utf-8 -*-
"""
Utilities related to managing threads used by nessai.
"""
import logging

import torch

logger = logging.getLogger(__name__)


def configure_threads(pytorch_threads=None):
    """Configure the number of threads available.

    This is necessary when using PyTorch on the CPU as by default it will use
    all available threads.

    Notes
    -----
    Uses ``torch.set_num_threads``. If pytorch threads is None but other
    arguments are specified then the value is inferred from them.

    Parameters
    ----------
    pytorch_threads: int, optional
        Maximum number of threads for PyTorch on CPU. If None, pytorch will
        use all available threads.
    """
    if pytorch_threads:
        logger.debug(
            f"Setting maximum number of PyTorch threads to {pytorch_threads}"
        )
        torch.set_num_threads(pytorch_threads)



# -*- coding: utf-8 -*-
"""
Base reparameterisation
"""
import logging

import numpy as np

logger = logging.getLogger(__name__)


class Reparameterisation:
    """
    Base object for reparameterisations.

    Parameters
    ----------
    parameters : str or list
        Name of parameters to reparameterise.
    prior_bounds : list, dict or None
        Prior bounds for the parameter(s).
    """

    _update_bounds = False
    has_prior = False
    has_prime_prior = False
    requires_prime_prior = False
    requires_bounded_prior = False
    prior_bounds = None
    prime_prior_bounds = None

    def __init__(self, parameters=None, prior_bounds=None):
        if not isinstance(parameters, (str, list)):
            raise TypeError("Parameters must be a str or list.")

        self.parameters = (
            [parameters] if isinstance(parameters, str) else parameters.copy()
        )

        if isinstance(prior_bounds, (list, tuple, np.ndarray)):
            if len(prior_bounds) == 2:
                prior_bounds = {self.parameters[0]: np.asarray(prior_bounds)}
            else:
                raise RuntimeError("Prior bounds got a list of len > 2")
        elif prior_bounds is None:
            if self.requires_bounded_prior:
                raise RuntimeError(
                    f"Reparameterisation {self.name} requires prior bounds!"
                )
            else:
                self.prior_bounds = None
        elif not isinstance(prior_bounds, dict):
            raise TypeError(
                "Prior bounds must be a dict, tuple, list or numpy array"
                " of len 2 or None."
            )

        if prior_bounds is not None:
            if set(self.parameters) - set(prior_bounds.keys()):
                raise RuntimeError(
                    "Mismatch between parameters and prior bounds: "
                    f"{set(self.parameters)}, {set(prior_bounds.keys())}"
                )
            self.prior_bounds = {
                p: np.asarray(b) for p, b in prior_bounds.items()
            }
        else:
            logger.debug(f"No prior bounds for {self.name}")

        if self.requires_bounded_prior:
            is_finite = np.isfinite(
                [pb for pb in self.prior_bounds.values()]
            ).all()
            if not is_finite:
                raise RuntimeError(
                    f"Reparameterisation {self.name} requires finite prior "
                    f"bounds. Received: {self.prior_bounds}"
                )

        self.prime_parameters = [p + "_prime" for p in self.parameters]
        self.requires = []
        logger.debug(f"Initialised reparameterisation: {self.name}")

    @property
    def name(self):
        """Unique name of the reparameterisations"""
        return (
            self.__class__.__name__.lower() + "_" + "_".join(self.parameters)
        )

    def reparameterise(self, x, x_prime, log_j):
        """
        Apply the reparameterisation to convert from x-space to x'-space.

        Parameters
        ----------
        x : structured array
            Array of inputs
        x_prime : structured array
            Array to be update
        log_j : array_like
            Log jacobian to be updated

        Returns
        -------
        x, x_prime : structured arrays
            Update version of the x and x_prime arrays
        log_j : array_like
            Updated log Jacobian determinant
        """
        raise NotImplementedError

    def inverse_reparameterise(self, x, x_prime, log_j):
        """
        Apply the reparameterisation to convert from x-space
        to x'-space

        Parameters
        ----------
        x : structured array
            Array
        x_prime : structured array
            Array to be update
        log_j : array_like
            Log jacobian to be updated

        Returns
        -------
        x, x_prime : structured arrays
            Update version of the x and x_prime arrays
        log_j : array_like
            Updated log Jacobian determinant
        """
        raise NotImplementedError

    def update(self, x):
        """Update the reparameterisation given some points.

        Does nothing by default.
        """
        pass



# -*- coding: utf-8 -*-
"""Base nested sampler object"""
from abc import ABC, abstractmethod
import datetime
import logging
import os
import pickle
import time
from typing import Any, Callable, Optional, Union

from glasflow import __version__ as glasflow_version
import numpy as np
import torch

from .. import __version__ as version
from ..model import Model
from ..utils import safe_file_dump

logger = logging.getLogger(__name__)


class BaseNestedSampler(ABC):
    """Base nested sampler class.

    Parameters
    ----------
    model : :obj:`nessai.model.Model`
        User-defined model
    output : str, optional
        Path for the output. If not specified the current working directory
        is used.
    seed : int, optional
        Seed used to seed numpy and torch.
    checkpointing : bool, optional
        Boolean to toggle checkpointing, must be enabled to resume the sampler.
        If false the sampler is still saved at the end of sampling.
    checkpoint_interval : int
        The interval used for checkpointing. By default this is a time interval
        in seconds. If :code:`checkpoint_on_iteration=True` this corresponds to
        the number of iterations between checkpointing.
    checkpoint_on_iteration : bool
        If true the checkpointing interval is checked against the number of
        iterations
    checkpoint_callback : Callback
        Callback function to be used instead of the default function. The
        function will be called in the code:`checkpoint` method as:
        :code`checkpoint_callback(self)`.
    logging_interval : int, optional
        The interval in seconds used for periodic logging. If not specified,
        then periodic logging is disabled.
    log_on_iteration : bool
        If true logging will occur based on the iteration. If false logging
        will be periodic if `logging_interval` is set. In case where neither
        logging is enabled, `log_on_iteration` will be set to true with an
        interval of :code:`nlive`.
    resume_file : str, optional
        Name of the file the sampler will be saved to and resumed from.
    plot : bool, optional
        Boolean to enable or disable plotting.
    n_pool : int, optional
        Number of threads to when for creating the multiprocessing pool.
    pool : object
        User defined multiprocessing pool that will be used when evaluating
        the likelihood.
    """

    def __init__(
        self,
        model: Model,
        nlive: int,
        output: str = None,
        seed: int = None,
        checkpointing: bool = True,
        checkpoint_interval: int = 600,
        checkpoint_on_iteration: bool = False,
        checkpoint_callback: Optional[Callable] = None,
        logging_interval: int = None,
        log_on_iteration: bool = True,
        resume_file: str = None,
        plot: bool = True,
        n_pool: Optional[int] = None,
        pool: Optional[Any] = None,
    ):
        logger.info("Initialising nested sampler")

        self.info_enabled = logger.isEnabledFor(logging.INFO)
        self.debug_enabled = logger.isEnabledFor(logging.DEBUG)
        model.verify_model()
        self.n_pool = n_pool
        self.model = model
        self.model.configure_pool(pool=pool, n_pool=n_pool)

        self.nlive = nlive
        self.plot = plot
        self.checkpointing = checkpointing
        self.checkpoint_interval = checkpoint_interval
        self.checkpoint_on_iteration = checkpoint_on_iteration
        self.checkpoint_callback = checkpoint_callback
        if self.checkpoint_on_iteration:
            self._last_checkpoint = 0
        else:
            self._last_checkpoint = datetime.datetime.now()
        self.sampling_time = datetime.timedelta()
        self.sampling_start_time = datetime.datetime.now()
        self.iteration = 0
        self.finalised = False
        self.resumed = False

        self.configure_random_seed(seed)
        self.configure_output(output, resume_file=resume_file)
        self.configure_periodic_logging(logging_interval, log_on_iteration)

        self.live_points = None
        self.history = None

    @property
    def current_sampling_time(self):
        if self.finalised:
            return self.sampling_time
        else:
            return self.sampling_time + (
                datetime.datetime.now() - self.sampling_start_time
            )

    @property
    def likelihood_evaluation_time(self):
        """Current log-likelihood time"""
        return self.model.likelihood_evaluation_time

    @property
    def total_likelihood_evaluations(self):
        """Total number of likelihood evaluations"""
        return self.model.likelihood_evaluations

    likelihood_calls = total_likelihood_evaluations
    """Alias for :code:`total_likelihood_evaluations`"""

    @property
    @abstractmethod
    def posterior_effective_sample_size(self):
        """The effective sample size of the posterior distribution"""
        raise NotImplementedError()

    def configure_output(
        self, output: Union[str, None], resume_file: Union[str, None] = None
    ):
        """Configure the output folder

        Parameters
        ----------
        output : str
            Directory where the results will be stored
        resume_file : str, optional
            Specific file to use for checkpointing. If not specified the
            default is used (nested_sampler_resume.pkl)
        """
        if output is None:
            output = os.getcwd()
        if not os.path.exists(output):
            os.makedirs(output, exist_ok=True)

        if resume_file is None:
            resume_file = os.path.join(output, "nested_sampler_resume.pkl")
        else:
            resume_file = os.path.join(output, resume_file)

        self.output = output
        self.resume_file = resume_file

    def configure_random_seed(self, seed: Optional[int]):
        """Initialise the random seed.

        Parameters
        ----------
        seed : Optional[int]
            The random seed. If not specified, a random seed is generated.
        """
        if seed is None:
            logger.debug("Seed not specified, generating random seed")
            seed = np.random.randint(0, min(np.iinfo(int).max, 2**32 - 1))
        logger.debug(f"Setting random seed to {seed}")
        self.seed = seed
        np.random.seed(seed=self.seed)
        torch.manual_seed(self.seed)

    def configure_periodic_logging(self, logging_interval, log_on_iteration):
        """Configure the periodic logging.

        Parameters
        ----------
        logging_interval : int, optional
            The interval in seconds used for periodic logging. If not
            specified, then periodic logging is disabled.
        log_on_iteration : bool
            If true logging will occur based on the iteration. If false logging
            will be periodic if `logging_interval` is set. In case where
            neither logging is enabled, `log_on_iteration` will be set to true
            with an interval of :code:`nlive`.
        """
        self.logging_interval = logging_interval
        self.log_on_iteration = log_on_iteration
        if not self.logging_interval and not self.log_on_iteration:
            logger.warning(
                "All logging disabled. Enabling logging on iteration"
            )
            self.log_on_iteration = True
        if self.log_on_iteration:
            if self.logging_interval is None:
                self.logging_interval = self.nlive
            self._last_log = 0
        else:
            self._last_log = time.time()

    @abstractmethod
    def log_state(self):
        raise NotImplementedError()

    def periodically_log_state(self):
        """Log the state of the sampler.

        Calls :code:`log_state` if the elapsed interval in time (or iterations)
        is more than the specified interval.
        """
        if self.log_on_iteration:
            if not (self.iteration - self._last_log) >= self.logging_interval:
                return
            else:
                self._last_log = self.iteration
        else:
            now = time.time()
            if not (now - self._last_log) >= self.logging_interval:
                return
            else:
                self._last_log = now
        self.log_state()

    def initialise_history(self) -> None:
        """Initialise the dictionary to store history"""
        if self.history is None:
            logger.debug("Initialising history dictionary")
            self.history = dict(
                likelihood_evaluations=[],
                sampling_time=[],
                checkpoint_iterations=[],
            )
        else:
            logger.debug("History dictionary already initialised")

    def update_history(self) -> None:
        """Update the history dictionary"""
        self.history["likelihood_evaluations"].append(
            self.total_likelihood_evaluations
        )
        self.history["sampling_time"].append(
            self.current_sampling_time.total_seconds()
        )

    def checkpoint(
        self,
        periodic: bool = False,
        force: bool = False,
        save_existing: bool = True,
    ):
        """Checkpoint the classes internal state.

        Parameters
        ----------
        periodic : bool
            Indicates if the checkpoint is regular periodic checkpointing
            or forced by a signal. If forced by a signal, it will show up on
            the state plot.
        force : bool
            Force the sampler to checkpoint.
        save_existing : bool
            If True, the previous checkpoint will be copied to `.old`.
            If False, the new checkpoint will override the previous file.
        """
        now = datetime.datetime.now()
        if not periodic:
            if self.history:
                self.history["checkpoint_iterations"] += [self.iteration]
            else:
                logger.warning(
                    "Could not log checkpoint iteration in the history"
                )
        elif force:
            pass
        else:
            if self.checkpoint_on_iteration:
                if (
                    self.iteration - self._last_checkpoint
                ) >= self.checkpoint_interval:
                    self._last_checkpoint = self.iteration
                else:
                    return
            else:
                if (
                    now - self._last_checkpoint
                ).total_seconds() >= self.checkpoint_interval:
                    self._last_checkpoint = now
                else:
                    return
        self.sampling_time += now - self.sampling_start_time
        logger.info("Checkpointing nested sampling")
        if self.checkpoint_callback:
            self.checkpoint_callback(self)
        else:
            safe_file_dump(
                self, self.resume_file, pickle, save_existing=save_existing
            )
        self.sampling_start_time = datetime.datetime.now()

    @classmethod
    def resume_from_pickled_sampler(
        cls,
        sampler: Any,
        model: Model,
        checkpoint_callback: Optional[Callable] = None,
    ):
        """Resume from pickle data.

        Parameters
        ----------
        data : Any
            Pickle data
        model : :obj:`nessai.model.Model`
            User-defined model
        checkpoint_callback : Optional[Callable]
            Checkpoint callback function. If not specified, the default method
            will be used.

        Returns
        -------
        Instance of BaseNestedSampler
        """
        logger.info(f"Resuming instance of {cls.__name__}")
        model.likelihood_evaluations += (
            sampler._previous_likelihood_evaluations
        )
        model.likelihood_evaluation_time += datetime.timedelta(
            seconds=sampler._previous_likelihood_evaluation_time
        )
        sampler.model = model
        sampler.resumed = True
        sampler.checkpoint_callback = checkpoint_callback
        return sampler

    @classmethod
    def resume(cls, filename: str, model: Model, **kwargs):
        """Resumes the interrupted state from a checkpoint pickle file.

        Parameters
        ----------
        filename : str
            Pickle file to resume from
        model : :obj:`nessai.model.Model`
            User-defined model

        Returns
        -------
        obj
            Instance of BaseNestedSampler
        """
        logger.info(f"Resuming {cls.__name__} from {filename}")
        with open(filename, "rb") as f:
            sampler = pickle.load(f)
        return cls.resume_from_pickled_sampler(sampler, model, **kwargs)

    @abstractmethod
    def nested_sampling_loop(self):
        raise NotImplementedError()

    def close_pool(self, code=None):
        """Close the multiprocessing pool."""
        self.model.close_pool(code=code)

    def get_result_dictionary(self):
        """Return a dictionary that contains results.

        Only includes version, seed and sampling time. Child classes should
        call this method and add to the dictionary.
        """
        d = dict()
        d["version"] = version
        d["glasflow_version"] = glasflow_version
        d["seed"] = self.seed
        d["sampling_time"] = self.sampling_time.total_seconds()
        d["total_likelihood_evaluations"] = self.model.likelihood_evaluations
        d["likelihood_evaluation_time"] = (
            self.likelihood_evaluation_time.total_seconds()
        )
        d["history"] = self.history
        if hasattr(self.model, "truth"):
            d["truth"] = self.model.truth
        return d

    def __getstate__(self):
        d = self.__dict__
        exclude = {"model", "proposal", "checkpoint_callback"}
        state = {k: d[k] for k in d.keys() - exclude}
        state["_previous_likelihood_evaluations"] = d[
            "model"
        ].likelihood_evaluations
        state["_previous_likelihood_evaluation_time"] = d[
            "model"
        ].likelihood_evaluation_time.total_seconds()
        return state



# -*- coding: utf-8 -*-
"""Test the version number."""
from importlib import reload
from unittest.mock import patch

try:
    from importlib.metadata import PackageNotFoundError

    LIB = "importlib.metadata"
except ImportError:
    from importlib_metadata import PackageNotFoundError

    LIB = "importlib_metadata"

import nessai
import pytest


@pytest.fixture(autouse=True)
def reload_nessai():
    """Make sure nessai is reloaded after these tests"""
    original_version = nessai.__version__
    yield
    reload(nessai)
    assert nessai.__version__ == original_version


def test_nessai_version():
    """Assert the correct version is set"""
    with patch(f"{LIB}.version", return_value="1.2.3"):
        reload(nessai)
    assert nessai.__version__ == "1.2.3"


def test_nessai_version_package_not_found():
    """Assert the version is unknown if the package is not installed."""
    with patch(f"{LIB}.version", side_effect=PackageNotFoundError):
        reload(nessai)
    assert nessai.__version__ == "unknown"



# -*- coding: utf-8 -*-
"""
Test the ToCartesian reparameterisation.
"""
import numpy as np
import pytest
from unittest.mock import create_autospec, patch

from nessai.livepoint import numpy_array_to_live_points
from nessai.reparameterisations import ToCartesian
from nessai.utils.testing import assert_structured_arrays_equal


@pytest.fixture
def reparam():
    return create_autospec(ToCartesian)


@pytest.fixture
def x():
    return numpy_array_to_live_points(np.array([[0.0], [1.0], [2.0]]), ["x"])


@pytest.mark.parametrize("mode", ["duplicate", "half", "split"])
@pytest.mark.parametrize("scale", [0.5, 2.0])
def test_init(reparam, mode, scale):
    prior_bounds = {"x": [0.0, 1.0]}
    """Test the init method"""

    def side_effect(scale, prior_bounds=None):
        reparam.scale = scale
        reparam.parameters = ["x", "y"]
        reparam.prior_bounds = prior_bounds

    with patch(
        "nessai.reparameterisations.Angle.__init__", side_effect=side_effect
    ) as super_init:
        ToCartesian.__init__(
            reparam, mode=mode, scale=scale, prior_bounds=prior_bounds
        )
    assert reparam.mode == mode
    assert reparam._zero_bound is False
    assert reparam._k == 1.0
    super_init.assert_called_once_with(scale=scale, prior_bounds=prior_bounds)


def test_init_invalid_mode(reparam):
    """Test the init method with an invalid mode"""
    with patch("nessai.reparameterisations.Angle.__init__"), pytest.raises(
        RuntimeError
    ) as excinfo:
        ToCartesian.__init__(reparam, mode="double")
    assert "Unknown mode" in str(excinfo.value)


def test_rescale_angle_split(reparam, x):
    """Test method for rescaling the 'angle'.

    For ToCartesian this is the rescaling applied to the parameter.
    """
    reparam.parameters = ["x"]
    reparam.prior_bounds = {"x": [0, 2]}
    reparam.mode = "split"
    reparam.scale = np.pi
    expected_angle = np.array([0.0, -np.pi / 2, np.pi])

    x_prime = numpy_array_to_live_points(np.zeros([3, 1]), ["x_prime"])
    log_j = np.zeros(3)

    with patch("numpy.random.choice", return_value=np.array([1])):
        angle, x_out, x_prime_out, log_j_out = ToCartesian._rescale_angle(
            reparam, x, x_prime, log_j, compute_radius=False
        )

    np.testing.assert_array_equal(angle, expected_angle)
    assert_structured_arrays_equal(x_out, x)
    assert_structured_arrays_equal(x_prime_out, x_prime)
    np.testing.assert_equal(log_j_out, -np.log(2))


@pytest.mark.parametrize(
    "args",
    [
        {"mode": "split", "compute_radius": True},
        {"mode": "duplicate", "compute_radius": False},
    ],
)
def test_rescale_angle_duplicate_or_compute_radius(reparam, x, args):
    """Test method for rescaling the 'angle'.

    For ToCartesian this is the rescaling applied to the parameter.
    """
    reparam.parameters = ["x"]
    reparam.prior_bounds = {"x": [0, 2]}
    reparam.mode = args["mode"]
    reparam.scale = np.pi

    expected_angle = np.array([0.0, np.pi / 2, np.pi, 0.0, -np.pi / 2, -np.pi])

    x_prime = numpy_array_to_live_points(np.zeros([3, 1]), ["x_prime"])
    log_j = np.zeros(3)

    with patch("numpy.random.choice", return_value=np.array([1])):
        angle, x_out, x_prime_out, log_j_out = ToCartesian._rescale_angle(
            reparam, x, x_prime, log_j, compute_radius=args["compute_radius"]
        )

    np.testing.assert_array_equal(angle, expected_angle)
    assert_structured_arrays_equal(x_out, np.concatenate([x, x]))
    assert_structured_arrays_equal(
        x_prime_out, np.concatenate([x_prime, x_prime])
    )
    np.testing.assert_equal(log_j_out, -np.log(2))


def test_inverse_rescale_angle(reparam):
    """Test the inverse method for rescaling the 'angle'."""
    reparam.parameters = ["x"]
    reparam.prior_bounds = {"x": [0, 2]}

    x = numpy_array_to_live_points(np.array([[0.0], [-0.5], [1.0]]), ["x"])
    expected_x = x.copy()
    expected_x["x"] = np.array([0.0, 1.0, 2.0])

    x_prime = numpy_array_to_live_points(
        np.array([0.0, 1.0, 3.0]), ["x_prime"]
    )
    log_j = np.zeros(3)

    x_out, x_prime_out, log_j_out = ToCartesian._inverse_rescale_angle(
        reparam, x, x_prime, log_j
    )

    assert_structured_arrays_equal(x_out, expected_x)
    assert_structured_arrays_equal(x_prime_out, x_prime)
    np.testing.assert_equal(log_j_out, np.log(2))



# -*- coding: utf-8 -*-
"""
Test the base proposal class.
"""
import logging
import pickle
import numpy as np
import pytest
from unittest.mock import MagicMock, Mock, create_autospec

from nessai.livepoint import numpy_array_to_live_points
from nessai.proposal.base import Proposal


class DummyProposal(Proposal):
    def draw(self, old_param):
        pass


@pytest.fixture
def proposal():
    return create_autospec(Proposal)


def test_init(proposal):
    """Test the init method"""
    model = Mock()
    Proposal.__init__(proposal, model)
    assert proposal.model is model
    assert proposal.training_count == 0
    assert proposal._checked_population is True
    assert proposal.population_time.total_seconds() == 0


def test_init_with_draw():
    """Assert an error is raised if `draw` is not implemented."""
    model = Mock()
    with pytest.raises(TypeError) as excinfo:
        Proposal(model)
    assert "class Proposal with" in str(excinfo.value)


def test_initialised(proposal):
    """Test the initialised property"""
    proposal._initialised = True
    val = Proposal.initialised.__get__(proposal)
    assert val is True


@pytest.mark.parametrize("val", [True, False])
def test_initialised_setter(proposal, val):
    """Test the setter for initialised."""
    Proposal.initialised.__set__(proposal, val)
    assert proposal._initialised is val


def test_initialise(proposal):
    """Test the initialise method"""
    Proposal.initialise(proposal)
    assert proposal.initialised is True


def test_evaluate_likelihoods(proposal):
    """Assert the correct method is called"""
    samples = numpy_array_to_live_points(np.array([[1], [2]]), ["x"])
    proposal.samples = samples
    proposal.model = MagicMock()
    proposal.model.batch_evaluate_log_likelihood = MagicMock(
        return_value=[1, 2]
    )
    Proposal.evaluate_likelihoods(proposal)
    proposal.model.batch_evaluate_log_likelihood.assert_called_once_with(
        samples
    )


def test_draw(proposal):
    """Assert an error is raised."""
    with pytest.raises(NotImplementedError):
        Proposal.draw(proposal, None)


def test_train(proposal, caplog):
    """Test the train method."""
    caplog.set_level(logging.INFO)
    Proposal.train(proposal, 1, plot=True)
    assert "This proposal method cannot be trained" in str(caplog.text)


def test_resume(proposal):
    """Test the resume method."""
    model = Mock()
    Proposal.resume(proposal, model)
    assert proposal.model is model


def test_getstate(proposal):
    """Test the get state method called by pickle."""
    proposal.model = Mock()
    state = Proposal.__getstate__(proposal)
    assert "model" not in list(state.keys())


@pytest.mark.integration_test
def test_pickling(model):
    """Test pickling the proposal"""
    proposal = DummyProposal(model)
    d = pickle.dumps(proposal)
    out = pickle.loads(d)
    assert hasattr(out, "model") is False



# -*- coding: utf-8 -*-
"""Transform for use in normalising flows."""
from warnings import warn
from glasflow.nflows.transforms import LULinear as BaseLULinear


class LULinear(BaseLULinear):
    """Wrapper for LULinear from nflows that works with CUDA.

    The original implementation has a bug that prevents use with CUDA. See
    https://github.com/bayesiains/nflows/pull/38 for details.

    This should be removed if the bug is fixed in nflows.
    """

    msg = (
        "`nessai.flows.transforms.LULinear` is deprecated and will be removed "
        "in a future release. Use `glasflow.nflows.transforms.LULinear` "
        "instead. "
    )
    warn(msg, FutureWarning)



# -*- coding: utf-8 -*-
"""
Testing the plotting functions.
"""
import os
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pytest
from unittest.mock import MagicMock, patch

from nessai import plot
from nessai import config
from nessai.livepoint import numpy_array_to_live_points


@pytest.fixture()
def live_points(model):
    """Set of live points"""
    return model.new_point(N=100)


@pytest.fixture()
def live_points_1(model):
    """Second set of live points"""
    return model.new_point(N=200)


@pytest.fixture()
def nested_samples(live_points):
    return np.sort(live_points, order="logL")


@pytest.fixture(autouse=True)
def auto_close_figures():
    """Automatically close all figures after each test"""
    yield
    plt.close("all")


@pytest.mark.parametrize("line_styles", [True, False])
def test_nessai_style_enabled(line_styles):
    """Assert the style is applied with config.plotting.disable_style=False

    Tests with `line_styles` True and False
    """

    def func(a, b):
        return a + b

    with patch("nessai.plot.config.plotting.disable_style", False), patch(
        "seaborn.axes_style"
    ) as mock_style, patch("matplotlib.rc_context") as mock_rc:
        out = plot.nessai_style(line_styles=line_styles)(func)(1, 2)
    assert out == 3
    mock_style.assert_called_with("ticks")
    mock_rc.assert_called_once()
    d = mock_rc.call_args[0][0]["axes.prop_cycle"].by_key()
    if line_styles:
        d["linestyle"] == config.plotting.line_styles
        d["color"] == config.plotting.line_colours
    else:
        assert "linestyle" not in d


def test_nessai_style_disabled():
    """
    Assert the style isn't applied with config.plotting.disable_style=True
    """

    def func(a, b):
        return a + b

    with patch("nessai.plot.config.plotting.disable_style", True), patch(
        "seaborn.axes_style"
    ) as mock_style:
        out = plot.nessai_style(func)(1, 2)
    assert out == 3
    mock_style.assert_not_called()


@pytest.mark.parametrize("line_styles", [True, False])
@pytest.mark.integration_test
def test_nessai_style_integration(line_styles):
    """Assert the line colours and styles are applied correctly"""

    def func():
        return (
            plt.rcParams["axes.prop_cycle"].by_key().get("color"),
            plt.rcParams["axes.prop_cycle"].by_key().get("linestyle", None),
        )

    colours, line_styles = plot.nessai_style(line_styles=line_styles)(func)()
    assert colours == config.plotting.line_colours
    if line_styles:
        assert line_styles == config.plotting.line_styles
    else:
        assert line_styles is None

    # Assert rcParams are still set to the defaults
    defaults = mpl.rcParamsDefault
    # Set backend manually
    defaults["backend"] = plt.rcParams["backend"]
    assert plt.rcParams == mpl.rcParamsDefault


@pytest.mark.parametrize("bounds", [None, True])
def test_plot_live_points_bounds(live_points, bounds, model):
    """Test generating a plot for a set of live points."""
    if bounds:
        bounds = model.bounds
    fig = plot.plot_live_points(live_points, bounds=bounds)
    assert fig is not None
    plt.close()


@pytest.mark.parametrize("c", [None, "x"])
def test_plot_live_points_hue(live_points, c, model):
    """Test generating a plot for a set of live points with a hue."""
    fig = plot.plot_live_points(live_points, c=c)
    assert fig is not None
    plt.close()


def test_plot_live_points_constant_hue(live_points):
    """Test to make sure that constant hue is handled correctly"""
    live_points["logL"] = np.ones(live_points.size)
    fig = plot.plot_live_points(live_points, c="logL")
    assert fig is not None
    plt.close()


@pytest.mark.parametrize("save", [False, True])
def test_plot_live_points_save(live_points, save, model, tmpdir):
    """Test generating a plot for a set of live points and saving it."""
    if save:
        filename = os.path.join(tmpdir, "lp.png")
    else:
        filename = None
    plot.plot_live_points(live_points, filename=filename)
    plt.close()


def test_plot_live_points_1d():
    """Test generating the live points plot for one parameter"""
    live_points = np.random.randn(100).view([("x", "f8")])
    fig = plot.plot_live_points(live_points)
    assert fig is not None
    plt.close()


def test_plot_live_points_with_nans(live_points):
    """Assert a plot is made if one of parameters is all NaNs"""
    live_points["x"] = np.nan
    fig = plot.plot_live_points(live_points)
    assert fig is not None
    plt.close()


@pytest.mark.parametrize("parameters", [None, ["x", "y"]])
def test_plot_1d_comparison_unstructured(parameters):
    """Test plotting live points in arrays are not structured."""
    live_points = np.random.randn(10, 2)
    plot.plot_1d_comparison(
        live_points, convert_to_live_points=True, parameters=parameters
    )
    plt.close()


def test_plot_1d_comparison_unstructured_missing_flag():
    """Test plotting live points in arrays are not structured."""
    live_points = np.random.randn(10, 2)
    with pytest.raises(RuntimeError) as excinfo:
        plot.plot_1d_comparison(live_points, convert_to_live_points=False)

    assert "not structured array" in str(excinfo.value)


@pytest.mark.parametrize("parameters", [None, ["x", "y"]])
def test_plot_1d_comparison_parameters(parameters, live_points, live_points_1):
    """Test generating a 1d comparison plot"""
    plot.plot_1d_comparison(live_points, live_points_1, parameters=parameters)
    plt.close()


@pytest.mark.parametrize("save", [False, True])
def test_plot_1d_comparison_save(save, live_points, live_points_1, tmpdir):
    """Test generating a 1d comparison plot"""
    if save:
        filename = os.path.join(tmpdir, "comp.png")
    else:
        filename = None
    plot.plot_1d_comparison(live_points, live_points_1, filename=filename)
    plt.close()


@pytest.mark.parametrize("labels", [None, True])
def test_plot_1d_comparison_labels(labels, live_points, live_points_1, model):
    """Test generating a 1d comparison plot"""
    if labels:
        labels = model.names
    plot.plot_1d_comparison(live_points, live_points_1, labels=labels)
    plt.close()


def test_plot_1d_comparison_invalid_labels_list(live_points, live_points_1):
    """Assert an error is raised if the colours list is a different length."""
    with pytest.raises(ValueError) as excinfo:
        plot.plot_1d_comparison(
            live_points, live_points_1, labels=["r", "g", "b"]
        )
    assert "Length of labels list must match " in str(excinfo.value)


@pytest.mark.parametrize("bounds", [None, True])
def test_plot_1d_comparison_bounds(bounds, live_points, live_points_1, model):
    """Test generating a 1d comparison plot"""
    if bounds:
        bounds = model.bounds
    plot.plot_1d_comparison(live_points, live_points_1, bounds=bounds)
    plt.close()


def test_plot_1d_comparison_1d():
    """Test generating a 1d comparison plot with only one parameter"""
    l1 = np.random.randn(100).view([("x", "f8")])
    l2 = np.random.randn(100).view([("x", "f8")])
    plot.plot_1d_comparison(l1, l2)
    plt.close()


def test_plot_1d_comparison_nans(live_points):
    """Assert parameters containing only NaNs don't raise an error"""
    live_points["logL"] = np.nan
    live_points["logP"] = np.nan
    plot.plot_1d_comparison(live_points)
    plt.close()


def test_plot_1d_comparison_infinite_var(live_points, caplog):
    """Test generating a 1d comparirson with a variable that is not finite."""
    live_points["y"] = np.inf * np.ones(live_points.size)
    plot.plot_1d_comparison(live_points)
    plt.close()
    assert "No finite points for y, skipping." in str(caplog.text)


def test_plot_1d_comparison_infinite_var_comp(
    live_points, live_points_1, caplog
):
    """Test generating a 1d comparirson with a variable that is not finite but
    the second set of points contains finite values.
    """
    live_points["y"] = np.inf * np.ones(live_points.size)
    plot.plot_1d_comparison(live_points, live_points_1)
    plt.close()
    assert "No finite points for y, skipping." not in str(caplog.text)


def test_plot_1d_comparison_one_valid(live_points, live_points_1, caplog):
    """Test generating a 1d comparison plot with one invalid"""
    live_points["logL"] = np.random.rand(live_points.size)
    live_points_1["logL"] = np.nan
    plot.plot_1d_comparison(live_points, live_points_1, parameters=["logL"])
    plt.close()
    assert "Could not plot legend" in caplog.text


def test_plot_1d_comparison_last_invalid(live_points, live_points_1):
    """Test generating a 1d comparison plot when the last parameter has NANs"""
    live_points["logL"] = np.random.rand(live_points.size)
    live_points_1["logL"] = np.nan
    plot.plot_1d_comparison(
        live_points, live_points_1, parameters=["x", "logL"]
    )
    plt.close()


@pytest.mark.parametrize("colours", [None, ["r", "g"]])
def test_plot_1d_comparison_colours(colours, live_points, live_points_1):
    """Test generating a 1d comparison plot with only one parameter"""
    plot.plot_1d_comparison(live_points, live_points_1, colours=colours)
    plt.close()


def test_plot_1d_comparison_more_colours(model):
    """Test generating a 1d comparirson when comparing more than 10 sets
    of live points.

    The default colour palettes all use 10 colours.
    """
    points = [model.new_point(N=10) for _ in range(12)]
    plot.plot_1d_comparison(*points)
    plt.close()


def test_plot_1d_comparison_invalid_colours_list(live_points, live_points_1):
    """Assert an error is raised if the colours list is a different length."""
    with pytest.raises(ValueError) as excinfo:
        plot.plot_1d_comparison(
            live_points, live_points_1, colours=["r", "g", "b"]
        )
    assert "Length of colours list must match " in str(excinfo.value)


@pytest.mark.parametrize("plot_breakdown", [False, True])
def test_plot_indices_breakdown(plot_breakdown):
    """Test plotting insertion indices with and without breakdown"""
    nlive = 100
    indices = np.random.randint(0, nlive, 1000)
    plot.plot_indices(indices, nlive=nlive, plot_breakdown=plot_breakdown)
    plt.close()


@pytest.mark.parametrize("save", [False, True])
def test_plot_indices_save(save, tmpdir):
    """Test plotting insertion indices with and without saving"""
    nlive = 100
    indices = np.random.randint(0, nlive, 1000)
    if save:
        filename = os.path.join(tmpdir, "indices.png")
    else:
        filename = None
    plot.plot_indices(indices, nlive=nlive, filename=filename)
    plt.close()


def test_plot_indices_no_indices():
    """Test to ensure plotting does not fail if indices are empty"""
    plot.plot_indices([], nlive=100)
    plt.close()


@pytest.mark.parametrize("sym", [False, True])
def test_plot_loss(sym):
    """Test function for plotting the loss"""
    epoch = 10
    if sym:
        history = dict(loss=-np.arange(epoch), val_loss=-np.arange(epoch))
    else:
        history = dict(loss=np.arange(epoch), val_loss=np.arange(epoch))
    plot.plot_loss(epoch, history)
    plt.close()


@pytest.mark.parametrize("save", [False, True])
def test_plot_loss_save(save, tmpdir):
    """Test function for plotting the loss"""
    epoch = 10
    history = dict(loss=np.arange(epoch), val_loss=np.arange(epoch))
    if save:
        filename = os.path.join(tmpdir, "loss.png")
    else:
        filename = None
    plot.plot_loss(epoch, history, filename=filename)
    plt.close()


@pytest.mark.parametrize("save", [False, True])
def test_trace_plot_save(nested_samples, save, tmpdir):
    """Test trace plot generation saving."""
    log_x = np.linspace(-10, 0, nested_samples.size)
    if save:
        filename = tmpdir + "trace.png"
    else:
        filename = None
    plot.plot_trace(log_x, nested_samples, filename=filename)
    plt.close()


def test_trace_plot_1d(nested_samples):
    """Test trace plot with only one parameter"""
    log_x = np.linspace(-10, 0, nested_samples.size)
    plot.plot_trace(log_x, nested_samples[["x"]])
    plt.close()


def test_trace_plot_unstructured():
    """
    Test to check that trace_plot raises an error when the nested samples
    are not a structured array.
    """
    log_x = np.linspace(-10, 0, 100)
    nested_samples = np.random.randn(log_x.size, 2)
    with pytest.raises(TypeError) as excinfo:
        plot.plot_trace(log_x, nested_samples)
    plt.close()
    assert "structured array" in str(excinfo.value)


@pytest.mark.parametrize(
    "labels", [None, ["x", "y"] + config.livepoints.non_sampling_parameters]
)
def test_trace_plot_labels(nested_samples, labels):
    """Test trace plot generation with labels."""
    log_x = np.linspace(-10, 0, nested_samples.size)
    plot.plot_trace(log_x, nested_samples, labels=labels)
    plt.close()


def test_trace_plot_labels_error(nested_samples):
    """Test to ensure error is raised if labels are incompatible"""
    log_x = np.linspace(-10, 0, nested_samples.size)
    with pytest.raises(
        RuntimeError, match=r"List of labels is the wrong length \(3\)"
    ):
        plot.plot_trace(log_x, nested_samples, labels=["1", "2", "3"])


def test_trace_plot_parameters(nested_samples):
    """Assert the parameters arguments works"""
    log_x = np.linspace(-10, 0, nested_samples.size)
    plot.plot_trace(log_x, nested_samples, parameters=["x"])
    plt.close()


def test_trace_plot_live_points(nested_samples):
    n = nested_samples.size // 2
    nested_samples, live_points = nested_samples[:n], nested_samples[n:]
    log_x = np.linspace(-10, -5, nested_samples.size)
    log_x_live_points = np.linspace(-5, 0, nested_samples.size)
    plot.plot_trace(
        log_x,
        nested_samples,
        live_points=live_points,
        log_x_live_points=log_x_live_points,
    )


def test_trace_plot_kwargs(nested_samples):
    """Assert the kwargs are passed to the plotting function."""
    log_x = np.linspace(-10, 0, nested_samples.size)
    mock_axes = MagicMock()
    mock_axes.plot = MagicMock()
    with patch("matplotlib.pyplot.subplots", return_value=(None, mock_axes)):
        plot.plot_trace(log_x, nested_samples, marker="^", parameters=["x"])
    mock_axes.plot.call_args[0][1] == dict(marker="^")
    plt.close()


def test_trace_plot_save_error(caplog, nested_samples):
    """Assert a warning is printed if the figure cannot be saved"""
    with patch.object(
        mpl.figure.Figure,
        "savefig",
        side_effect=ValueError,
    ) as mock_save:
        plot.plot_trace(
            np.arange(nested_samples.size),
            nested_samples,
            filename="trace.png",
        )
    mock_save.assert_called_once()
    assert "Could not save trace plot" in str(caplog.text)


def test_histogram_plot():
    """Test the basic histogram plot"""
    x = np.random.randn(100)
    with patch("matplotlib.pyplot.hist") as mocked_hist:
        plot.plot_histogram(x, bins=10, density=True, label="test")
    mocked_hist.assert_called_once_with(
        x, bins=10, density=True, histtype="step"
    )


def test_histogram_plot_save(tmpdir):
    """Test to make sure the figure is saved."""
    x = np.random.randn(100)
    filename = os.path.join(tmpdir, "hist.png")
    fig = plot.plot_histogram(x, filename=filename)
    assert fig is None
    assert os.path.exists(filename)


def test_corner_plot(live_points):
    """Test the corner plot."""
    fig = plot.corner_plot(live_points)
    assert fig is not None


def test_corner_plot_check_inputs(live_points):
    """Assert corner.corner is called with the correct inputs"""
    with patch("corner.corner") as mock_corner:
        plot.corner_plot(live_points, test_kwarg="a")

    assert len(mock_corner.call_args[0]) == 1
    kwargs = mock_corner.call_args[1]
    assert "truths" in kwargs
    assert "labels" in kwargs
    assert "test_kwarg" in kwargs
    assert "color" in kwargs


@pytest.mark.parametrize(
    "labels",
    [["x", "y"], ["x", "y"] + config.livepoints.non_sampling_parameters],
)
def test_corner_plot_w_labels(live_points, labels):
    """Test the corner plot with labels"""
    plot.corner_plot(live_points, labels=labels)


@pytest.mark.parametrize(
    "truths",
    [[0, 0], [0, 0, None, None, None], {"x": 0, "y": 0}, np.array([0, 0])],
)
def test_corner_plot_w_truths(live_points, truths):
    """Test the corner plot with truths"""
    plot.corner_plot(live_points, truths=truths)


def test_corner_plot_w_exclude(live_points):
    """Test the parameter is excluded"""
    fig = plot.corner_plot(live_points, exclude=["y"])
    assert len(fig.axes) == 1


def test_corner_plot_w_include(live_points):
    """Test the parameter is included"""
    fig = plot.corner_plot(live_points, include=["x"])
    assert len(fig.axes) == 1


def test_corner_plot_w_include_and_labels(live_points):
    """Test the parameter is included and the labels do not raise an error"""
    fig = plot.corner_plot(live_points, include=["x"], labels=["x_0"])
    assert len(fig.axes) == 1


@pytest.mark.parametrize("truths", [[1], {"x": 1, "y": 1}])
def test_corner_plot_w_include_and_truths(live_points, truths):
    """Test the parameter is included and the truths do not raise an error"""
    fig = plot.corner_plot(live_points, include=["x"], truths=truths)
    assert len(fig.axes) == 1


def test_corner_plot_w_include_and_truths_error(live_points):
    """Assert an error is raised when the number truths do not match the
    number of parameters
    """
    with pytest.raises(ValueError, match=r"truths does not match .*"):
        plot.corner_plot(live_points, include=["x"], truths=[1, 1])


def test_corner_plot_all_nans(caplog, live_points):
    """Test how NaNs are handled.

    Should skip a parameter will al NaNs. In this case this should also include
    all of the non-sampling parameters since they are either NaNs or have zero
    dynamic range (it).
    """
    live_points["x"] = np.nan
    fig = plot.corner_plot(live_points)
    assert fig is not None
    assert (
        str(["x"] + config.livepoints.non_sampling_parameters) in caplog.text
    )


def test_corner_plot_save(tmpdir, live_points):
    """Assert the corner plot is saved."""
    filename = os.path.join(tmpdir, "corner.png")
    fig = plot.corner_plot(live_points, filename=filename)
    assert fig is None
    assert os.path.exists(filename)


def test_corner_plot_fields_exclude_error(live_points):
    """Assert an error is raised if include and exclude are both specified"""
    with pytest.raises(ValueError) as excinfo:
        plot.corner_plot(live_points, include=["x"], exclude=["logL"])
    assert "Cannot specify both `include` and `exclude`" in str(excinfo.value)


def test_corner_plot_save_error(caplog, live_points):
    """Assert a warning is printed if the figure cannot be saved"""
    with patch.object(
        mpl.figure.Figure,
        "savefig",
        side_effect=ValueError,
    ) as mock_save:
        plot.corner_plot(live_points, filename="corner.png")
    mock_save.assert_called_once()
    assert "Could not save corner plot" in str(caplog.text)


@pytest.mark.integration_test
@pytest.mark.parametrize("dims", [10, 100, 200])
def test_plot_trace_large_dims(tmp_path, dims):
    """Test producing a trace plot with a large number of dimensions."""
    n = 10_000
    names = [f"x_{i}" for i in range(dims)]
    log_x = np.arange(n)
    x = numpy_array_to_live_points(np.random.randn(n, dims), names)

    fig = plot.plot_trace(log_x, x)
    fig.savefig(tmp_path / "trace.png")



from nessai.samplers.importancesampler import ImportanceNestedSampler as INS
import pytest
from unittest.mock import MagicMock


@pytest.mark.parametrize(
    "criterion, tolerance, stop_any, reached",
    [
        ([5.0], [0.0], False, False),
        ([-1.0], [0.0], False, True),
        ([5.0, 1.5], [0.0, 1], False, False),
        ([5.0, 0.5], [0.0, 1], False, False),
        ([-1.0, 0.5], [0.0, 1], False, True),
        ([5.0, 1.5], [0.0, 1], True, False),
        ([5.0, 0.5], [0.0, 1], True, True),
        ([-1.0, 0.5], [0.0, 1], True, True),
    ],
)
def test_reached_tolerance(ins, criterion, tolerance, stop_any, reached):
    ins.criterion = criterion
    ins.tolerance = tolerance
    ins._stop_any = stop_any
    assert INS.reached_tolerance.__get__(ins) is reached


def test_compute_stopping_criterion_fractional_error(ins):
    ins.stopping_criterion = ["fractional_error"]
    ins.iteration = 0
    ins.tolerance = 0.0
    ins.state = MagicMock()
    ins.state.evidence = 1.0
    ins.state.evidence_error = 0.1
    assert INS.compute_stopping_criterion(ins) == [0.1]



# -*- coding: utf-8 -*-
"""
Tests for threading related utilities.
"""
from unittest.mock import patch

from nessai.utils.threading import configure_threads


def test_configure_threads():
    """Test configuring the threads"""
    with patch("torch.set_num_threads") as mock:
        configure_threads(pytorch_threads=1)
    mock.assert_called_once_with(1)


def test_configure_threads_none():
    """Test configuring the threads.

    Assert `set_num_threads` it not called if pytorch_threads is None.
    """
    with patch("torch.set_num_threads") as mock:
        configure_threads(pytorch_threads=None)
    mock.assert_not_called()



# -*- coding: utf-8 -*-
"""Test the proposal utilities"""
from nessai.gw.proposal import GWFlowProposal
from nessai.proposal.flowproposal import FlowProposal
from nessai.proposal.utils import check_proposal_kwargs
import pytest


def test_class_inheritance():
    """Assert classes that inherit from another class work"""

    class A:
        def __init__(self, a, b=1, c=2) -> None:
            self.a = a
            self.b = b
            self.c = c

    class B(A):
        def __init__(self, a, b=3, d=4, **kwargs) -> None:
            super().__init__(a, b, **kwargs)
            self.d = d

    kwargs = dict(b=10, c=20, d=30)
    kwargs_out = check_proposal_kwargs(B, kwargs)
    assert kwargs_out == dict(b=10, c=20, d=30)


def test_kwargs_all_okay(caplog):
    """Assert the kwargs are returned if they all correspond to the class"""
    caplog.set_level("DEBUG")
    kwargs = dict(poolsize=100, volume_fraction=0.9)
    expected = kwargs.copy()
    out = check_proposal_kwargs(GWFlowProposal, kwargs)
    assert out == expected
    assert "All keyword arguments match the proposal class" in str(caplog.text)


@pytest.mark.parametrize("ProposalClass", [FlowProposal, GWFlowProposal])
def test_remove_kwargs(ProposalClass, caplog):
    """Assert kwargs for a different proposal are removed"""
    kwargs = dict(poolsize=100, volume_fraction=0.9, augment_dims=1)
    expected = kwargs.copy()
    expected.pop("augment_dims")
    out = check_proposal_kwargs(ProposalClass, kwargs)
    assert out == expected
    assert "Removing unused keyword arguments" in str(caplog.text)


def test_check_kwargs_strict():
    """Assert kwargs for a different proposal raise an error if strict=True"""
    kwargs = dict(poolsize=100, volume_fraction=0.9, augment_dims=1)
    with pytest.raises(
        RuntimeError,
        match="Keyword arguments contain unknown keys: {'augment_dims'}",
    ):
        check_proposal_kwargs(FlowProposal, kwargs, strict=True)


def test_check_kwargs_error():
    """Assert an error is raised if a keyword argument is not known."""
    with pytest.raises(
        RuntimeError, match=r"Unknown kwargs for FlowProposal: {'not_a_kwarg'}"
    ):
        check_proposal_kwargs(FlowProposal, dict(not_a_kwarg=None))



# -*- coding: utf-8 -*-
"""
Utilities for the test suite.
"""
import numpy as np

from ..model import Model


class IntegrationTestModel(Model):
    """Complete nessai model for use with integration tests"""

    def __init__(self, dims=2):
        self.bounds = {f"x_{i}": [-5.0, 5.0] for i in range(dims)}
        self.names = list(self.bounds.keys())

    def log_prior(self, x):
        log_p = np.log(self.in_bounds(x), dtype="float")
        for n in self.names:
            log_p -= np.log(self.bounds[n][1] - self.bounds[n][0])
        return log_p

    def log_likelihood(self, x):
        return np.sum(-0.5 * (self.unstructured_view(x) ** 2), axis=-1)

    def to_unit_hypercube(self, x):
        x_out = x.copy()
        for n in self.names:
            x_out[n] = (x[n] - self.bounds[n][0]) / np.ptp(self.bounds[n])
        return x_out

    def from_unit_hypercube(self, x):
        x_out = x.copy()
        for n in self.names:
            x_out[n] = np.ptp(self.bounds[n]) * x[n] + self.bounds[n][0]
        return x_out


def assert_structured_arrays_equal(x, y, atol=0.0, rtol=0.0):
    """Assert structured arrays are equal.

    Supports NaNs by checking each field individually.

    Parameters
    ----------
    x : np.ndarray
        Array to check.
    y : np.ndarray
        Array to compare to.
    atol : float
        The absolute tolerance parameter. See the numpy documentation for all
        :code:`numpy.allclose`.
    rtol : float
        The relative tolerance parameter. See the numpy documentation for all
        :code:`numpy.allclose`.

    Raises
    ------
    AssertionError
        If dtype or values in each field are not equal.
    """
    x = np.asanyarray(x)
    y = np.asanyarray(y)
    if x.dtype != y.dtype:
        raise AssertionError(
            f"""Structured array dtypes are different:

            Expected: {x.dtype}
            Actual: {y.dtype}
            """
        )

    valid = {f: False for f in x.dtype.names}
    for field in valid.keys():
        valid[field] = np.allclose(
            x[field], y[field], equal_nan=True, atol=atol, rtol=rtol
        )

    if not all(valid.values()):
        mismatched = [k for k, v in valid.items() if v is False]
        msg = f"""
        Arrays are not equal.

        Mismatched fields: {mismatched}
        """
        for f in mismatched:
            msg += f"""
            Field: {f}
            Expected: {y[f]}
            Actual: {x[f]}
            """
        raise AssertionError(msg)



# -*- coding: utf-8 -*-
"""
Test the base reparameterisation.
"""
import numpy as np
from numpy.testing import assert_equal
import pytest
from unittest.mock import create_autospec

from nessai.reparameterisations import Reparameterisation


@pytest.fixture()
def reparam():
    return create_autospec(Reparameterisation)


@pytest.mark.parametrize("name", ["x1", ["x1"]])
@pytest.mark.parametrize("prior_bounds", [[0, 1], (0, 1), {"x1": [0, 1]}])
def test_init(name, prior_bounds):
    """Test the init method with the allowed types of inputs"""
    reparam = Reparameterisation(parameters=name, prior_bounds=prior_bounds)
    assert reparam.parameters == ["x1"]
    assert reparam.prime_parameters == ["x1_prime"]
    assert_equal(reparam.prior_bounds, {"x1": np.array([0, 1])})


def test_init_infinite_bounds():
    """Test the init method with infinite prior bounds"""
    reparam = Reparameterisation(
        parameters=["x", "y"], prior_bounds={"x": [0, 1], "y": [0, np.inf]}
    )
    assert reparam.parameters == ["x", "y"]
    assert reparam.prime_parameters == ["x_prime", "y_prime"]
    assert_equal(reparam.prior_bounds["x"], [0, 1])
    assert_equal(reparam.prior_bounds["y"], [0, np.inf])


def test_infinite_bounds_error():
    """Test to ensure infinite prior bounds raise an error.

    Only applies if `requires_bounded_prior` is True.
    """

    class TestReparam(Reparameterisation):
        requires_bounded_prior = True

    with pytest.raises(RuntimeError) as excinfo:
        TestReparam(
            parameters=["x", "y"], prior_bounds={"x": [0, 1], "y": [0, np.inf]}
        )
    assert "requires finite prior" in str(excinfo.value)


def test_no_prior_bounds():
    """Test not providing prior bounds."""

    class TestReparam(Reparameterisation):
        requires_bounded_prior = False

    reparam = TestReparam(parameters=["x", "y"])
    assert reparam.prior_bounds is None


def test_no_prior_bounds_error():
    """Test missing prior bounds error.

    Only applies if `requires_bounded_prior` is True.
    """

    class TestReparam(Reparameterisation):
        requires_bounded_prior = True

    with pytest.raises(RuntimeError) as excinfo:
        TestReparam(parameters=["x", "y"])
    assert "requires prior bounds" in str(excinfo.value)


def test_parameters_error():
    with pytest.raises(TypeError) as excinfo:
        Reparameterisation(parameters={"x": [0, 1]})
    assert "Parameters must be a str or list" in str(excinfo.value)


def test_missing_bounds():
    with pytest.raises(RuntimeError) as excinfo:
        Reparameterisation(parameters=["x", "y"], prior_bounds={"x": [0, 1]})
    assert "Mismatch" in str(excinfo.value)


def test_incorrect_bounds_type():
    with pytest.raises(TypeError) as excinfo:
        Reparameterisation(parameters=["x", "y"], prior_bounds=1)
    assert "Prior bounds must be" in str(excinfo.value)


def test_incorrect_bounds_length():
    with pytest.raises(RuntimeError) as excinfo:
        Reparameterisation(parameters=["x", "y"], prior_bounds=[1, 2, 3])
    assert "Prior bounds got a list of len > 2" in str(excinfo.value)


def test_methods_not_implemented():
    """Test to ensure class fails if user does not define the methods"""
    reparam = Reparameterisation(parameters="x", prior_bounds=[0, 1])

    with pytest.raises(NotImplementedError):
        reparam.reparameterise(None, None, None)

    with pytest.raises(NotImplementedError):
        reparam.inverse_reparameterise(None, None, None)


def test_update(reparam):
    """Assert the default update method can be called and does not raised an
    error.
    """
    x = np.array((1, 2), dtype=[("x", "f8"), ("y", "f8")])
    Reparameterisation.update(reparam, x)



# -*- coding: utf-8 -*-
"""
Base objects for implementing normalising flows.
"""
from abc import ABC, abstractmethod
from torch.nn import Module


class BaseFlow(Module, ABC):
    """
    Base class for all normalising flows.

    If implementing flows using distributions and transforms see NFlow.
    """

    device = None

    def to(self, device):
        """Wrapper that stores the device before moving the flow"""
        self.device = device
        super().to(device)

    @abstractmethod
    def forward(self, x, context=None):
        """
        Apply the forward transformation and return samples in the
        latent space and the log-Jacobian determinant.

        Returns
        -------
        :obj:`torch.Tensor`
            Tensor of samples in the latent space
        :obj:`torch.Tensor`
            Tensor of log determinants of the Jacobian of the forward
            transformation
        """
        raise NotImplementedError()

    @abstractmethod
    def inverse(self, z, context=None):
        """
        Apply the inverse transformation and return samples in the
        data space and the log-Jacobian determinant.

        Returns
        -------
        :obj:`torch.Tensor`
            Tensor of samples in the data space
        :obj:`torch.Tensor`
            Tensor of log determinants of the Jacobian of the forward
            transformation
        """
        raise NotImplementedError()

    @abstractmethod
    def sample(self, n, context=None):
        """
        Generate n samples in the data space

        Returns
        -------
        :obj:`torch.Tensor`
            Tensor of samples in the data space
        """
        raise NotImplementedError()

    @abstractmethod
    def log_prob(self, x, context=None):
        """
        Compute the log probability for a set of samples in the data space

        Returns
        -------
        :obj:`torch.Tensor`
            Tensor of log probabilities of the samples
        """
        raise NotImplementedError()

    @abstractmethod
    def sample_latent_distribution(self, n, context=None):
        """Sample from the latent distribution."""
        raise NotImplementedError

    @abstractmethod
    def base_distribution_log_prob(self, z, context=None):
        """
        Computes the log probability of samples in the latent for
        the base distribution in the flow.

        Returns
        -------
        :obj:`torch.Tensor`
            Tensor of log probabilities of the latent samples
        """
        raise NotImplementedError()

    @abstractmethod
    def forward_and_log_prob(self, x, context=None):
        """
        Apply the forward transformation and compute the log probability
        of each sample

        Returns
        -------
        :obj:`torch.Tensor`
            Tensor of samples in the latent space
        :obj:`torch.Tensor`
            Tensor of log probabilities of the samples
        """
        raise NotImplementedError()

    @abstractmethod
    def sample_and_log_prob(self, n, context=None):
        """
        Generates samples from the flow, together with their log probabilities
        in the data space ``log p(x) = log p(z) + log|J|``.

        For flows, this is more efficient that calling ``sample`` and
        ``log_prob`` separately.

        Returns
        -------
        :obj:`torch.Tensor`
            Tensor of samples in the data space
        :obj:`torch.Tensor`
            Tensor of log probabilities of the samples
        """
        raise NotImplementedError()

    def finalise(self):
        """Finalise the flow after training.

        Will be called after training the flow and loading the best weights.
        For example, can be used to finalise the Monte Carlo estimate of the
        normalising constant used in a LARS based flow.

        By default does nothing and should be implemented by the user.
        """
        pass

    def end_iteration(self):
        """Update the model at the end of an iteration.

        Will be called between training and validation.

        By default does nothing and should be overridden by an class that
        inherit from this class.
        """
        pass

    @abstractmethod
    def freeze_transform(self):
        """Freeze the transform part of the flow.

        Must be implemented by the child class.
        """
        raise NotImplementedError()

    @abstractmethod
    def unfreeze_transform(self):
        """Unfreeze the transform part of the flow.

        Must be implemented by the child class.
        """
        raise NotImplementedError()


class NFlow(BaseFlow):
    """Base class for flow objects from glasflow.nflows.

    This replaces `Flow` from glasflow.nflows. It includes additional methods
    which are called in FlowModel.

    Parameters
    ----------
    transform : :obj: `glasflow.nflows.transforms.Transform`
        Object that applies the transformation, must have`forward` and
        `inverse` methods. See glasflow.nflows for more details.
    distribution : :obj: `glasflow.nflows.distributions.Distribution`
        Object the serves as the base distribution used when sampling
        and computing the log probability. Must have `log_prob` and
        `sample` methods. See glasflow.nflows for details
    """

    def __init__(self, transform, distribution):
        super().__init__()
        from glasflow.nflows.transforms import Transform
        from glasflow.nflows.distributions import Distribution

        if not isinstance(transform, Transform):
            raise TypeError(
                "transform must inherit from "
                "`glasflow.nflows.transforms.Transform`. "
                f"Got: {type(transform)}"
            )

        if not isinstance(distribution, Distribution):
            raise TypeError(
                "distribution must inherit from "
                "`glasflow.nflows.transforms.Distribution`. "
                f"Got: {type(transform)}"
            )

        self._transform = transform
        self._distribution = distribution

    def forward(self, x, context=None):
        """
        Apply the forward transformation and return samples in the latent
        space and log-Jacobian determinant.
        """
        return self._transform.forward(x, context=context)

    def inverse(self, z, context=None):
        """
        Apply the inverse transformation and return samples in the
        data space and log-Jacobian determinant (not log probability).
        """
        return self._transform.inverse(z, context=context)

    def sample(self, num_samples, context=None):
        """
        Produces N samples in the data space by drawing from the base
        distribution and the applying the inverse transform.

        Does NOT need to be specified by the user
        """
        noise = self._distribution.sample(num_samples)

        samples, _ = self._transform.inverse(noise, context=context)

        return samples

    def log_prob(self, inputs, context=None):
        """
        Computes the log probability of the inputs samples by apply the
        transform.

        Does NOT need to specified by the user
        """
        noise, logabsdet = self._transform(inputs, context=context)
        log_prob = self._distribution.log_prob(noise)
        return log_prob + logabsdet

    def sample_latent_distribution(self, n, context=None):
        if context is not None:
            raise NotImplementedError
        return self._distribution.sample(n)

    def base_distribution_log_prob(self, z, context=None):
        """
        Computes the log probability of samples in the latent for
        the base distribution in the flow.
        """
        return self._distribution.log_prob(z)

    def forward_and_log_prob(self, x, context=None):
        """
        Apply the forward transformation and compute the log probability
        of each sample

        Returns
        -------
        :obj:`torch.Tensor`
            Tensor of samples in the latent space
        :obj:`torch.Tensor`
            Tensor of log probabilities of the samples
        """
        z, log_J = self.forward(x, context=context)
        log_prob = self.base_distribution_log_prob(z)
        return z, log_prob + log_J

    def sample_and_log_prob(self, N, context=None):
        """
        Generates samples from the flow, together with their log probabilities
        in the data space ``log p(x) = log p(z) + log|J|``.

        For flows, this is more efficient that calling ``sample`` and
        ``log_prob`` separately.
        """
        z, log_prob = self._distribution.sample_and_log_prob(N)

        samples, logabsdet = self._transform.inverse(z, context=context)

        return samples, log_prob - logabsdet

    def finalise(self):
        """Finalise the flow after training.

        Checks if the base distribution or transform have finalise methods
        and calls them.
        """
        if hasattr(self._distribution, "finalise"):
            self._distribution.finalise()
        if hasattr(self._transform, "finalise"):
            self._transform.finalise()

    def end_iteration(self):
        """Update the model at the end of an iteration.

        Will be called between training and validation.
        """
        if hasattr(self._distribution, "end_iteration"):
            self._distribution.end_iteration()
        if hasattr(self._transform, "end_iteration"):
            self._transform.end_iteration()

    def freeze_transform(self):
        """Freeze the transform part of the flow"""
        self._transform.requires_grad_(False)

    def unfreeze_transform(self):
        """Unfreeze the transform part of the flow"""
        self._transform.requires_grad_(True)



# -*- coding: utf-8 -*-
"""Test compatibility with bilby"""
import numpy as np
import pytest


@pytest.mark.requires("bilby")
@pytest.mark.bilby_compatibility
@pytest.mark.slow_integration_test
def test_bilby_compatiblity(tmp_path):
    """Test compatibility with bilby"""
    import bilby

    outdir = tmp_path / "bilby_test"

    class GaussianLikelihood(bilby.Likelihood):
        def __init__(self):
            super().__init__(parameters={"x": None, "y": None})

        def log_likelihood(self):
            """Log-likelihood."""
            return -0.5 * (
                self.parameters["x"] ** 2.0 + self.parameters["y"] ** 2.0
            ) - np.log(2.0 * np.pi)

    priors = dict(
        x=bilby.core.prior.Uniform(-5, 5, "x"),
        y=bilby.core.prior.Uniform(-5, 5, "y"),
    )

    bilby.run_sampler(
        outdir=outdir,
        resume=False,
        plot=False,
        likelihood=GaussianLikelihood(),
        priors=priors,
        sampler="nessai",
        injection_parameters={"x": 0.0, "y": 0.0},
        analytic_priors=True,
        nlive=500,
        seed=1234,
        logging_interval=10,
        log_on_iteration=False,
    )



# -*- coding: utf-8 -*-
"""Tests for GW reparameterisations"""

import pytest
from nessai.gw.reparameterisations import (
    DeltaPhaseReparameterisation,
    DistanceReparameterisation,
    default_gw,
    get_gw_reparameterisation,
)
from nessai.utils.testing import assert_structured_arrays_equal
import numpy as np
from unittest.mock import MagicMock, patch, create_autospec

from nessai.livepoint import (
    dict_to_live_points,
    empty_structured_array,
)


@pytest.fixture
def distance_reparam():
    return create_autospec(DistanceReparameterisation)


@pytest.fixture
def delta_phase_reparam():
    return create_autospec(DeltaPhaseReparameterisation)


def test_get_gw_reparameterisation():
    """Test getting the gw reparameterisations.

    Assert the correct defaults are used.
    """
    expected = "out"
    with patch(
        "nessai.gw.reparameterisations.get_reparameterisation",
        return_value=expected,
    ) as base_fn:
        out = get_gw_reparameterisation("mass_ratio")
    assert out == expected
    base_fn.assert_called_once_with("mass_ratio", defaults=default_gw)


@pytest.mark.integration_test
def test_get_gw_reparameterisation_integration():
    """Integration test for get_gw_reparameterisation"""
    reparam, _ = get_gw_reparameterisation("distance")
    assert reparam is DistanceReparameterisation


@pytest.mark.parametrize(
    "has_conversion, has_jacobian, has_prime_prior, requires_prime_prior",
    [
        (False, True, False, False),
        (False, False, False, False),
        (True, False, True, True),
        (True, True, True, False),
    ],
)
def test_distance_reparameterisation_init(
    distance_reparam,
    has_conversion,
    has_jacobian,
    has_prime_prior,
    requires_prime_prior,
):
    """Test the init method for the DistanceReparameterisation class.

    Tests the different combinations of conversions and jacobians.
    """
    prior = "uniform-comoving-volume"
    parameter = "parameter"
    prior_bounds = {"parameter": [10.0, 100.0]}

    distance_reparam.detect_edges_kwargs = {}
    distance_reparam.requires_prime_prior = False
    distance_reparam.update_prime_prior_bounds = MagicMock()
    mock_converter = MagicMock()
    mock_converter.has_conversion = has_conversion
    mock_converter.has_jacobian = has_jacobian
    mock_converter_class = MagicMock(return_value=mock_converter)

    with patch(
        "nessai.gw.reparameterisations.get_distance_converter",
        return_value=mock_converter_class,
    ) as converter_fn:
        DistanceReparameterisation.__init__(
            distance_reparam,
            parameters=parameter,
            prior_bounds=prior_bounds,
            prior=prior,
        )

    converter_fn.assert_called_once_with(prior)
    mock_converter_class.assert_called_once_with(d_min=10.0, d_max=100.0)
    assert distance_reparam.has_prime_prior is has_prime_prior
    assert distance_reparam.requires_prime_prior is requires_prime_prior
    # If the reparam includes a conversion, the prime priors should be updated
    assert distance_reparam.update_prime_prior_bounds.called is has_conversion


def test_distance_reparameterisation_n_parameters_error(distance_reparam):
    """Assert an error is raised in more than one parameter is given."""
    prior = "uniform-comoving-volume"
    parameters = ["x", "y"]
    prior_bounds = {"x": [10.0, 100.0], "y": [10.0, 100.0]}
    with pytest.raises(RuntimeError) as excinfo:
        DistanceReparameterisation.__init__(
            distance_reparam,
            parameters=parameters,
            prior_bounds=prior_bounds,
            prior=prior,
        )

    assert "DistanceReparameterisation only supports one parameter" in str(
        excinfo.value
    )


def test_delta_phase_init(delta_phase_reparam):
    """Assert the parent method is called and the parameters are set."""
    parameters = "phase"
    prior_bounds = {"phase": [0, 6.28]}
    with patch(
        "nessai.gw.reparameterisations.Reparameterisation.__init__"
    ) as mock:
        DeltaPhaseReparameterisation.__init__(
            delta_phase_reparam,
            parameters=parameters,
            prior_bounds=prior_bounds,
        )
    mock.assert_called_once_with(
        parameters=parameters, prior_bounds=prior_bounds
    )
    assert delta_phase_reparam.requires == ["psi", "theta_jn"]
    assert delta_phase_reparam.prime_parameters == ["delta_phase"]


def test_delta_phase_reparameterise(delta_phase_reparam):
    """Assert the correct value is returned"""
    delta_phase_reparam.parameters = ["phase"]
    delta_phase_reparam.prime_parameters = ["delta_phase"]

    x = dict(phase=1.0, theta_jn=0.0, psi=0.5)
    x_prime = dict(delta_phase=np.nan, theta_jn=0.0, psi=0.5)
    log_j = 0

    (
        x_out,
        x_prime_out,
        log_j_out,
    ) = DeltaPhaseReparameterisation.reparameterise(
        delta_phase_reparam, x, x_prime, log_j
    )
    assert x_out == x
    assert x_prime_out["delta_phase"] == 1.5
    assert log_j_out == 0


def test_delta_phase_inverse_reparameterise(delta_phase_reparam):
    """Assert the correct value is returned"""
    delta_phase_reparam.parameters = ["phase"]
    delta_phase_reparam.prime_parameters = ["delta_phase"]

    x = dict(phase=np.nan, theta_jn=0.0, psi=0.5)
    x_prime = dict(delta_phase=0.5, theta_jn=0.0, psi=0.5)
    log_j = 0

    (
        x_out,
        x_prime_out,
        log_j_out,
    ) = DeltaPhaseReparameterisation.inverse_reparameterise(
        delta_phase_reparam, x, x_prime, log_j
    )
    assert x_prime_out == x_prime
    assert x["phase"] == 0.0
    assert log_j_out == 0


@pytest.mark.integration_test
def test_delta_phase_inverse_invertible():
    """Assert the reparameterisation is invertible"""
    n = 10
    parameters = ["phase"]
    prior_bounds = {"phase": [0.0, 2 * np.pi]}
    reparam = DeltaPhaseReparameterisation(
        parameters=parameters, prior_bounds=prior_bounds
    )
    x = dict_to_live_points(
        {
            "phase": np.random.uniform(0, 2 * np.pi, n),
            "psi": np.random.uniform(0, np.pi, n),
            "theta_jn": np.random.uniform(0, np.pi, n),
        }
    )
    x_prime = empty_structured_array(
        n, names=["delta_phase", "theta_jn", "psi"]
    )
    x_prime["psi"] = x["psi"]
    x_prime["theta_jn"] = x["theta_jn"]
    log_j = np.zeros(n)
    x_f, x_prime_f, log_j_f = reparam.reparameterise(x, x_prime, log_j)

    x_in = x_f.copy()
    x_in["phase"] = np.nan

    assert_structured_arrays_equal(x_f, x)
    np.testing.assert_array_equal(log_j_f, log_j)
    x_i, x_prime_i, log_j_i = reparam.inverse_reparameterise(
        x_in, x_prime_f.copy(), log_j_f.copy()
    )
    assert_structured_arrays_equal(x_prime_i, x_prime_f)
    assert_structured_arrays_equal(x_i, x, rtol=1e-10)
    np.testing.assert_array_equal(log_j_i, log_j_f)
